import traceback
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Type, cast

from confluent_kafka import TopicPartition, cimpl
from django.utils import timezone

from django_kafka.conf import settings
from django_kafka.consumer import Consumer, Topics
from django_kafka.dead_letter.topic import DeadLetterTopicProducer
from django_kafka.retry.header import RetryHeader
from django_kafka.retry.topic import RetryTopicConsumer

if TYPE_CHECKING:
    from django_kafka.topic import TopicConsumer


class RetryTopics(Topics):
    def __init__(self, group_id: str, *topic_consumers: "TopicConsumer"):
        super().__init__(
            *(
                RetryTopicConsumer(group_id=group_id, topic_consumer=t)
                for t in topic_consumers
            ),
        )


class RetryConsumer(Consumer):
    topics: RetryTopics
    resume_times: dict[TopicPartition, datetime]

    def __init__(self):
        super().__init__()
        self.resume_times = {}

    @classmethod
    def build(cls, consumer_cls: Type["Consumer"]) -> Optional[Type["RetryConsumer"]]:
        """Generates RetryConsumer subclass linked to consumer class retryable topics"""
        retryable_tcs = [t for t in consumer_cls.topics if t.retry_settings]
        if not retryable_tcs:
            return None

        group_id = consumer_cls.build_config()["group.id"]

        return type[RetryConsumer](
            f"{consumer_cls.__name__}Retry",
            (cls,),
            {
                "config": {
                    **getattr(cls, "config", {}),
                    "group.id": f"{group_id}.retry",
                },
                "topics": RetryTopics(group_id, *retryable_tcs),
            },
        )

    @classmethod
    def build_config(cls):
        return {
            **super().build_config(),
            **settings.RETRY_CONSUMER_CONFIG,
            **getattr(cls, "config", {}),
        }

    def retry_msg(self, msg: cimpl.Message, exc: Exception) -> bool:
        rt_consumer = cast(RetryTopicConsumer, self.get_topic_consumer(msg))
        return rt_consumer.producer_for(msg).retry(exc)

    def dead_letter_msg(self, msg: cimpl.Message, exc: Exception):
        rt_consumer = cast(RetryTopicConsumer, self.get_topic_consumer(msg))
        DeadLetterTopicProducer(group_id=rt_consumer.group_id, msg=msg).produce_for(
            header_message=str(exc),
            header_detail=traceback.format_exc(),
        )

    def pause_partition(self, msg, until: datetime):
        """pauses the partition and stores the resumption time"""
        tp = TopicPartition(msg.topic(), msg.partition(), msg.offset())
        self.seek(tp)  # seek back to message offset, so it is re-polled on unpause
        self.pause([tp])
        self.resume_times[tp] = until

    def resume_ready_partitions(self):
        """resumes any partitions that were paused"""
        now = timezone.now()
        for tp, until in list(self.resume_times.items()):
            if now < until:
                continue
            self.resume([tp])
            del self.resume_times[tp]

    def poll(self):
        self.resume_ready_partitions()
        return super().poll()

    def process_message(self, msg: cimpl.Message):
        retry_time = RetryHeader.get_retry_time(msg.headers())
        if retry_time and retry_time > timezone.now():
            self.pause_partition(msg, retry_time)
            return
        super().process_message(msg)

    def stop(self):
        self.resume_times = {}
        super().stop()



import logging
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Type

from confluent_kafka.serialization import (
    MessageField,
    SerializationContext,
    Serializer,
    StringSerializer,
)

from django_kafka import kafka
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.retry.settings import RetrySettings

if TYPE_CHECKING:
    from confluent_kafka import cimpl

logger = logging.getLogger(__name__)


class TopicProducer(ABC):
    key_serializer: Type[Serializer] = StringSerializer
    value_serializer: Type[Serializer] = StringSerializer

    @property
    @abstractmethod
    def name(self) -> str:
        """Define Kafka topic producing topic name"""

    def __init__(self):
        super().__init__()
        if self.name.startswith("^"):
            raise DjangoKafkaError(f"TopicProducer name cannot be regex: {self.name}")

    def get_key_serializer(self, **kwargs):
        return self.key_serializer(**kwargs)

    def get_value_serializer(self, **kwargs):
        return self.value_serializer(**kwargs)

    def context(
        self,
        field: MessageField,
        headers: Optional[list] = None,
    ) -> SerializationContext:
        return SerializationContext(self.name, field, headers=headers)

    def serialize(
        self,
        value,
        field: MessageField,
        headers: Optional[list] = None,
        **kwargs,
    ):
        if field == MessageField.VALUE:
            serializer = self.get_value_serializer(**kwargs)
            return serializer(value, self.context(MessageField.VALUE, headers))

        if field == MessageField.KEY:
            serializer = self.get_key_serializer(**kwargs)
            return serializer(value, self.context(MessageField.KEY, headers))

        raise DjangoKafkaError(f"Unsupported serialization field {field}.")

    def produce(self, value: any, **kwargs):
        key_serializer_kwargs = kwargs.pop("key_serializer_kwargs", {}) or {}
        value_serializer_kwargs = kwargs.pop("value_serializer_kwargs", {}) or {}
        headers = kwargs.get("headers")

        if "key" in kwargs:
            kwargs["key"] = self.serialize(
                kwargs["key"],
                MessageField.KEY,
                headers,
                **key_serializer_kwargs,
            )

        kafka.producer.produce(
            self.name,
            self.serialize(
                value,
                MessageField.VALUE,
                headers,
                **value_serializer_kwargs,
            ),
            **kwargs,
        )
        kafka.producer.poll(0)  # stops producer on_delivery callbacks buffer overflow


class TopicConsumer(ABC):
    key_deserializer: Type[Serializer] = StringSerializer
    value_deserializer: Type[Serializer] = StringSerializer

    retry_settings: Optional["RetrySettings"] = None

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Defines Kafka topic consumer subscription name.

        Regexp pattern subscriptions are supported by prefixing the name with "^".
        """

    def is_regex(self) -> bool:
        """Returns if the subscription name is regex based"""
        return self.name.startswith("^")

    def matches(self, topic_name: str) -> bool:
        if self.is_regex():
            return bool(re.search(self.name, topic_name))
        return self.name == topic_name

    def get_key_deserializer(self, **kwargs):
        return self.key_deserializer(**kwargs)

    def get_value_deserializer(self, **kwargs):
        return self.value_deserializer(**kwargs)

    def context(
        self,
        field: MessageField,
        headers: Optional[list] = None,
    ) -> SerializationContext:
        return SerializationContext(self.name, field, headers=headers)

    def deserialize(
        self,
        value,
        field: MessageField,
        headers: Optional[list] = None,
        **kwargs,
    ):
        if field == MessageField.VALUE:
            deserializer = self.get_value_deserializer(**kwargs)
            return deserializer(value, self.context(MessageField.VALUE, headers))

        if field == MessageField.KEY:
            deserializer = self.get_key_deserializer(**kwargs)
            return deserializer(value, self.context(MessageField.KEY, headers))

        raise DjangoKafkaError(f"Unsupported deserialization field {field}.")

    def consume(self, msg: "cimpl.Message"):
        """Implement message processing"""
        raise NotImplementedError


class Topic(TopicConsumer, TopicProducer, ABC):
    """Combined Topic class for when topic consume and produce names are the same"""



from typing import TYPE_CHECKING, Optional, Type

from django.utils import timezone

if TYPE_CHECKING:
    from django_kafka.topic import TopicConsumer


class RetrySettings:
    def __init__(
        self,
        max_retries: int,
        delay: int,
        backoff: bool = False,
        include: Optional[list[Type[Exception]]] = None,
        exclude: Optional[list[Type[Exception]]] = None,
    ):
        """
        :param max_retries: maximum number of retry attempts
        :param delay: delay (seconds)
        :param backoff: exponential backoff
        :param include: exception types to retry for
        :param exclude: exception types to exclude from retry
        """
        if max_retries <= 0:
            raise ValueError("max_retries must be greater than zero")
        if delay <= 0:
            raise ValueError("delay must be greater than zero")
        if include is not None and exclude is not None:
            raise ValueError("cannot specify both include and exclude")

        self.max_retries = max_retries
        self.delay = delay
        self.backoff = backoff
        self.include = include
        self.exclude = exclude

    def __call__(self, topic_cls: Type["TopicConsumer"]):
        topic_cls.retry_settings = self
        return topic_cls

    def attempts_exceeded(self, attempt):
        return attempt > self.max_retries

    def should_retry(self, exc: Exception) -> bool:
        if self.include is not None:
            return any(isinstance(exc, e) for e in self.include)
        if self.exclude is not None:
            return not any(isinstance(exc, e) for e in self.exclude)

        return True

    def get_retry_timestamp(self, attempt: int) -> str:
        delay = self.delay * 2 ** (attempt - 1) if self.backoff else self.delay
        return str(timezone.now().timestamp() + delay)



from typing import Optional

import requests
from requests.auth import AuthBase
from urllib3.util import Retry
from requests.adapters import HTTPAdapter

from django_kafka.exceptions import DjangoKafkaError


class KafkaConnectSession(requests.Session):
    def __init__(
        self,
        host: str,
        auth: Optional[tuple | AuthBase] = None,
        retry: dict = None,
        timeout: int = None,
    ):
        super().__init__()
        self.auth = auth
        self.host = host
        self.timeout = timeout
        if retry:
            self.mount_retry_adapter(retry)

    def mount_retry_adapter(self, retry: dict):
        self.mount(self.host, HTTPAdapter(max_retries=Retry(**retry)))

    def request(self, method, url, *args, **kwargs) -> requests.Response:
        kwargs.setdefault("timeout", self.timeout)
        return super().request(method, f"{self.host}{url}", *args, **kwargs)


class KafkaConnectClient:
    """
    https://kafka.apache.org/documentation/#connect_rest
    https://docs.confluent.io/platform/current/connect/references/restapi.html
    """

    def __init__(
        self,
        host: str,
        auth: Optional[tuple | AuthBase] = None,
        retry: dict = None,
        timeout: int = None,
    ):
        self._requests = KafkaConnectSession(host, auth, retry, timeout)

    def update_or_create(self, connector_name: str, config: dict):
        return self._requests.put(f"/connectors/{connector_name}/config", json=config)

    def delete(self, connector_name: str):
        return self._requests.delete(f"/connectors/{connector_name}")

    def validate(self, config: dict):
        if not config.get('connector.class'):
            raise DjangoKafkaError("'connector.class' config is required for validation.")

        connector_class_name = config.get("connector.class").rsplit(".", 1)[-1]
        response = self._requests.put(f"/connector-plugins/{connector_class_name}/config/validate", json=config)
        return response

    def connector_status(self, connector_name: str):
        """
        https://docs.confluent.io/platform/current/connect/references/restapi.html#get--connectors-(string-name)-status
        """
        return self._requests.get(f"/connectors/{connector_name}/status")



from datetime import datetime
from enum import StrEnum
from typing import Optional

from django.utils import timezone


class RetryHeader(StrEnum):
    MESSAGE = "RETRY_MESSAGE"
    TIMESTAMP = "RETRY_TIMESTAMP"

    @staticmethod
    def get_header(headers: list[tuple], header: "RetryHeader") -> Optional[str]:
        return next((v for k, v in headers if k == header), None)

    @staticmethod
    def get_retry_time(headers: Optional[list[tuple]]) -> Optional["datetime"]:
        """returns the retry time from the message headers"""
        if not headers:
            return None

        header = RetryHeader.get_header(headers, RetryHeader.TIMESTAMP)
        try:
            epoch = float(header)
        except (TypeError, ValueError):
            return None
        else:
            return datetime.fromtimestamp(epoch, tz=timezone.get_current_timezone())



import re
from typing import TYPE_CHECKING

from django_kafka.conf import settings
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.retry.header import RetryHeader
from django_kafka.serialization import NoOpSerializer
from django_kafka.topic import TopicConsumer, TopicProducer

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.retry.settings import RetrySettings


class RetryTopicProducer(TopicProducer):
    key_serializer = NoOpSerializer
    value_serializer = NoOpSerializer

    def __init__(
        self,
        group_id: str,
        retry_settings: "RetrySettings",
        msg: "cimpl.Message",
    ):
        self.settings = retry_settings
        self.group_id = group_id
        self.msg = msg
        self.attempt = self.get_next_attempt(msg.topic())
        super().__init__()

    @classmethod
    def suffix(cls):
        return settings.RETRY_TOPIC_SUFFIX

    @classmethod
    def pattern(cls):
        """returns regex pattern to match topics producing by this class"""
        return rf"{re.escape(cls.suffix())}\.([0-9]+)$"

    @classmethod
    def get_next_attempt(cls, topic_name: str) -> int:
        match = re.search(cls.pattern(), topic_name)
        attempt = int(match.group(1)) if match else 0
        return attempt + 1

    @property
    def name(self) -> str:
        topic = self.msg.topic()
        suffix = f"{self.suffix()}.{self.attempt}"

        if re.search(self.pattern(), topic):
            return re.sub(self.pattern(), suffix, topic)
        return f"{self.group_id}.{self.msg.topic()}.{suffix}"

    def retry(self, exc: Exception) -> bool:
        if not self.settings.should_retry(exc=exc):
            return False

        if self.settings.attempts_exceeded(attempt=self.attempt):
            return False

        self.produce(
            key=self.msg.key(),
            value=self.msg.value(),
            headers=[
                (RetryHeader.MESSAGE, str(exc)),
                (
                    RetryHeader.TIMESTAMP,
                    self.settings.get_retry_timestamp(self.attempt),
                ),
            ],
        )
        return True


class RetryTopicConsumer(TopicConsumer):
    key_deserializer = NoOpSerializer
    value_deserializer = NoOpSerializer

    def __init__(self, group_id: str, topic_consumer: TopicConsumer):
        if not topic_consumer.retry_settings:
            raise DjangoKafkaError(
                f"TopicConsumer {topic_consumer} is not marked for retry",
            )
        self.topic_consumer = topic_consumer
        self.group_id = group_id
        super().__init__()

    @property
    def name(self) -> str:
        """returns name as regex pattern matching all attempts on the group's topic"""
        group_id = re.escape(self.group_id)
        if self.topic_consumer.is_regex():
            topic_name = f"({self.topic_consumer.name.lstrip('^').rstrip('$')})"
        else:
            topic_name = re.escape(self.topic_consumer.name)

        return rf"^{group_id}\.{topic_name}\.{RetryTopicProducer.pattern()}"

    def consume(self, msg: "cimpl.Message"):
        self.topic_consumer.consume(msg)

    def producer_for(self, msg: "cimpl.Message") -> RetryTopicProducer:
        return RetryTopicProducer(
            group_id=self.group_id,
            retry_settings=self.topic_consumer.retry_settings,
            msg=msg,
        )



from unittest.mock import patch, Mock

from django.test import SimpleTestCase

from django_kafka.connect.client import KafkaConnectClient
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.connect.connector import Connector
from django_kafka.conf import settings


class MyConnector(Connector):
    config = {}


@patch("django_kafka.connect.client.KafkaConnectSession", new=Mock())
@patch.multiple("django_kafka.conf.settings", CONNECT_HOST="http://kafka-connect")
class ConnectorTestCase(SimpleTestCase):
    def test_name(self):
        class MyConnector2(Connector):
            name = 'custom-name'
            config = {}

        self.assertEqual(MyConnector.name, 'MyConnector')
        self.assertEqual(MyConnector2.name, 'custom-name')

    @patch("django_kafka.connect.connector.KafkaConnectClient", spec=True)
    def test_init_request_session(self, mock_client):
        connector = MyConnector()

        mock_client.assert_called_with(
            host=settings.CONNECT_HOST,
            auth=settings.CONNECT_AUTH,
            retry=settings.CONNECT_RETRY,
            timeout=settings.CONNECT_REQUESTS_TIMEOUT,
        )
        self.assertIsInstance(connector.client, KafkaConnectClient)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.delete.return_value": Mock(ok=True),
    })
    def test_delete_response_ok(self, mock_client):
        connector = MyConnector()

        result = connector.delete()

        self.assertIs(result, True)
        mock_client().delete.assert_called_once_with(connector.name)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.delete.return_value": Mock(ok=False, status_code=404),
    })
    def test_delete_404(self, mock_client):
        connector = MyConnector()
        result = connector.delete()

        self.assertIs(result, False)
        mock_client().delete.assert_called_once_with(connector.name)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.delete.return_value": Mock(ok=False),
    })
    def test_delete_response_not_ok_raise_error(self, mock_client):
        connector = MyConnector()

        with self.assertRaises(DjangoKafkaError):
            MyConnector().delete()

        mock_client().delete.assert_called_once_with(connector.name)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.submit.return_value": Mock(ok=True),
    })
    def test_submit_response_ok(self, mock_client):
        connector = MyConnector()

        result = connector.submit()

        mock_client().update_or_create.assert_called_once_with(connector.name, connector.config)
        response_json = mock_client().update_or_create().json()
        self.assertEqual(response_json, result)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.update_or_create.return_value": Mock(ok=False),
    })
    def test_submit_response_not_ok_raise_error(self, mock_client):
        connector = MyConnector()

        with self.assertRaises(DjangoKafkaError):
            connector.submit()

        mock_client().update_or_create.assert_called_once_with(connector.name, connector.config)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.validate.return_value": Mock(ok=True),
    })
    def test_is_valid_response_ok(self, mock_client):
        connector = MyConnector()

        result = connector.is_valid()

        mock_client().validate.assert_called_once_with(connector.config)
        response_ok = mock_client().validate().ok
        self.assertEqual(response_ok, result)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.validate.return_value": Mock(ok=False),
    })
    def test_is_valid_response_not_ok_no_error(self, mock_client):
        connector = MyConnector()

        result = connector.is_valid()

        mock_client().validate.assert_called_once_with(connector.config)
        response_ok = mock_client().validate().ok
        self.assertIs(response_ok, False)
        self.assertEqual(result, response_ok)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.validate.return_value": Mock(ok=False),
    })
    def test_is_valid_response_not_ok_raise_error(self, mock_client):
        connector = MyConnector()

        with self.assertRaises(DjangoKafkaError):
            connector.is_valid(raise_exception=True)

        mock_client().validate.assert_called_once_with(connector.config)

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.connector_status.return_value": Mock(
            ok=True,
            json=Mock(return_value={"connector": {"state": "RUNNING"}}),
        ),
    })
    def test_status_response_ok(self, mock_client):
        connector = MyConnector()

        result = connector.status()

        mock_client().connector_status.assert_called_once_with(connector.name)
        self.assertEqual(result, mock_client().connector_status().json()["connector"]["state"])

    @patch("django_kafka.connect.connector.KafkaConnectClient", **{
        "return_value.connector_status.return_value": Mock(ok=False),
    })
    def test_status_response_not_ok_raise_error(self, mock_client):
        connector = MyConnector()

        with self.assertRaises(DjangoKafkaError):
            connector.status()

        mock_client().connector_status.assert_called_once_with(connector.name)



from unittest.mock import patch

from django.test import override_settings, SimpleTestCase

from django_kafka import DjangoKafka
from django_kafka.conf import SETTINGS_KEY
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.producer import Producer
from django_kafka.registry import ConsumersRegistry


class DjangoKafkaTestCase(SimpleTestCase):
    def test_registry(self):
        kafka = DjangoKafka()

        # registry is there (registry itself is tested in a separate testcase)
        self.assertIsInstance(kafka.consumers, ConsumersRegistry)

    def test_producer(self):
        kafka = DjangoKafka()

        producer = kafka.producer
        # producer is there
        self.assertIsInstance(producer, Producer)
        # producer instance is the same (cached)
        self.assertIs(kafka.producer, producer)

    @patch("django_kafka.SchemaRegistryClient")
    def test_schema_client(self, mock_schema_registry_client):
        kafka = DjangoKafka()

        # exception when config is not provided
        with self.assertRaisesMessage(
            DjangoKafkaError,
            "`SCHEMA_REGISTRY` configuration is not defined.",
        ):
            # ruff: noqa: B018
            kafka.schema_client

        schema_registry_config = {"url": "http://schema-registry"}
        with override_settings(
            **{SETTINGS_KEY: {"SCHEMA_REGISTRY": schema_registry_config}},
        ):
            schema_client = kafka.schema_client
            # porper instance returned
            self.assertIs(schema_client, mock_schema_registry_client.return_value)
            # instantiated only once (cached)
            self.assertIs(kafka.schema_client, mock_schema_registry_client.return_value)
            # called with right args
            mock_schema_registry_client.assert_called_once_with(schema_registry_config)

    @patch("django_kafka.DjangoKafka.consumers")
    def test_run_consumer(self, mock_django_kafka_consumers):
        kafka = DjangoKafka()
        mock_consumer = mock_django_kafka_consumers["path.to.Consumer"]()

        kafka.run_consumer("path.to.Consumer")

        mock_consumer.run.assert_called_once()



import traceback
from contextlib import suppress
from unittest.mock import MagicMock, Mock, call, patch

from django.test import TestCase, override_settings

from django_kafka.conf import SETTINGS_KEY, settings
from django_kafka.consumer import Consumer, Topics


class StopWhileTrue:
    class Error(Exception):
        pass

    def __init__(self, stop_on):
        self.stop_on = stop_on

    def __call__(self, arg):
        if arg == self.stop_on:
            raise self.Error()


class ConsumerTestCase(TestCase):
    def test_group_id(self):
        class SomeConsumer(Consumer):
            config = {"group.id": "group_id"}

        consumer = SomeConsumer()

        self.assertEqual(consumer.group_id, "group_id")

    @patch(
        "django_kafka.consumer.Consumer.process_message",
        side_effect=StopWhileTrue(stop_on=False),
    )
    @patch(
        "django_kafka.consumer.ConfluentConsumer",
        **{
            # first iteration - no message (None)
            # second iteration - 1 non-empty message represented as True for testing
            # third iteration - 1 empty message represented as False for testing
            "return_value.poll.side_effect": [None, True, False],
        },
    )
    def test_run(self, mock_consumer_client, mock_process_message):
        class SomeConsumer(Consumer):
            topics = MagicMock()
            log_error = Mock()

        consumer = SomeConsumer()

        # hack to break infinite loop
        # `consumer.start` is using `while True:` loop which never ends
        # in order to test it we need to break it which is achievable with side_effect
        with suppress(StopWhileTrue.Error):
            consumer.run()

        # subscribed to the topics defined by consumer class
        mock_consumer_client.return_value.subscribe.assert_called_once_with(
            topics=consumer.topics.names,
        )
        # poll for message with the frequency defined by consumer class
        self.assertEqual(
            mock_consumer_client.return_value.poll.call_args_list,
            # 3 calls expected as we setup 3 iterations in the test
            [call(timeout=consumer.polling_freq)] * 3,
        )
        # process_message called twice as None message is ignored
        self.assertEqual(mock_process_message.call_args_list, [call(True), call(False)])

    @patch("django_kafka.consumer.Consumer.commit_offset")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message_success(self, mock_consumer_client, mock_commit_offset):
        class SomeConsumer(Consumer):
            topics = MagicMock()

        msg = Mock(error=Mock(return_value=False))

        consumer = SomeConsumer()

        consumer.process_message(msg)

        # check if msg has error before processing
        msg.error.assert_called_once_with()
        # Topic.consume called
        consumer.get_topic_consumer(msg.topic()).consume.assert_called_once_with(msg)
        # commit_offset triggered
        mock_commit_offset.assert_called_once_with(msg)

    @patch("django_kafka.consumer.Consumer.log_error")
    @patch("django_kafka.consumer.Consumer.commit_offset")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message_msg_error_logged(
        self,
        mock_consumer_client,
        mock_commit_offset,
        log_error,
    ):
        class SomeConsumer(Consumer):
            topics = MagicMock()

        msg = Mock(error=Mock(return_value=True))

        consumer = SomeConsumer()

        consumer.process_message(msg)

        # check if msg has error before processing
        msg.error.assert_called_once_with()
        # error handler was triggered
        log_error.assert_called_once_with(msg.error.return_value)
        # Topic.consume is not called
        consumer.topics[msg.topic()].consume.assert_not_called()
        # Consumer.commit_offset is not called
        mock_commit_offset.assert_not_called()

    @patch("django_kafka.consumer.Consumer.handle_exception")
    @patch("django_kafka.consumer.Consumer.commit_offset")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message_exception(
        self,
        mock_consumer_client,
        mock_commit_offset,
        handle_exception,
    ):
        topic_consume_side_effect = TypeError("test")
        topic_consumer = Mock(
            **{
                "name": "topic",
                "consume.side_effect": topic_consume_side_effect,
            },
        )
        msg = Mock(
            **{"topic.return_value": topic_consumer.name, "error.return_value": False},
        )

        class SomeConsumer(Consumer):
            topics = Topics(topic_consumer)

        consumer = SomeConsumer()

        consumer.process_message(msg)

        # check if msg has error before processing
        msg.error.assert_called_once_with()
        # Topic.consume was triggered
        topic_consumer.consume.assert_called_once_with(msg)
        # error handler was triggered on exception
        handle_exception.assert_called_once_with(msg, topic_consume_side_effect)
        # Consumer.commit_offset is not called
        mock_commit_offset.assert_called_once()

    def test_handle_exception(self):
        msg = Mock()

        class SomeConsumer(Consumer):
            topics = Topics()
            config = {"group.id": "group_id"}
            retry_msg = Mock(return_value=True)  # successful retry
            dead_letter_msg = Mock()
            log_error = Mock()

        consumer = SomeConsumer()
        exc = ValueError()

        consumer.handle_exception(msg, exc)

        consumer.retry_msg.assert_called_once_with(msg, exc)
        consumer.dead_letter_msg.assert_not_called()
        consumer.log_error.assert_not_called()

    def test_handle_exception__failed_retry(self):
        msg = Mock()

        class SomeConsumer(Consumer):
            topics = Topics()
            config = {"group.id": "group_id"}
            retry_msg = Mock(return_value=False)  # failed retry
            dead_letter_msg = Mock()
            log_error = Mock()

        consumer = SomeConsumer()
        exc = ValueError()

        consumer.handle_exception(msg, exc)

        consumer.retry_msg.assert_called_once_with(msg, exc)
        consumer.dead_letter_msg.assert_called_once_with(msg, exc)
        consumer.log_error.assert_called_once_with(exc)

    @patch("django_kafka.retry.topic.RetryTopicProducer")
    def test_retry_msg(self, mock_rt_producer_cls):
        mock_topic_consumer = Mock(retry_settings=Mock())
        mock_retry = mock_rt_producer_cls.return_value.retry
        msg_mock = Mock()

        class SomeConsumer(Consumer):
            topics = Topics()
            config = {"group.id": "group_id"}

        consumer = SomeConsumer()
        consumer.get_topic_consumer = Mock(return_value=mock_topic_consumer)
        exc = ValueError()

        retried = consumer.retry_msg(msg_mock, exc)

        mock_rt_producer_cls.assert_called_once_with(
            group_id=consumer.group_id,
            retry_settings=mock_topic_consumer.retry_settings,
            msg=msg_mock,
        )
        mock_retry.assert_called_once_with(exc=exc)
        self.assertEqual(retried, mock_retry.return_value)

    @patch("django_kafka.retry.topic.RetryTopicProducer")
    def test_retry_msg__no_retry(self, mock_rt_producer_cls):
        mock_topic_consumer = Mock(retry_settings=None)
        msg_mock = Mock()

        class SomeConsumer(Consumer):
            topics = Topics()
            config = {"group.id": "group_id"}

        consumer = SomeConsumer()
        consumer.get_topic_consumer = Mock(return_value=mock_topic_consumer)
        exc = ValueError()

        retried = consumer.retry_msg(msg_mock, exc)

        mock_rt_producer_cls.assert_not_called()
        self.assertEqual(retried, False)

    @patch("django_kafka.dead_letter.topic.DeadLetterTopicProducer")
    def test_dead_letter_msg(self, mock_dead_letter_topic_cls):
        mock_produce_for = mock_dead_letter_topic_cls.return_value.produce_for
        msg_mock = Mock()

        class SomeConsumer(Consumer):
            topics = Topics()
            config = {"group.id": "group_id"}

        consumer = SomeConsumer()
        exc = ValueError()

        consumer.dead_letter_msg(msg_mock, exc)

        mock_dead_letter_topic_cls.assert_called_once_with(
            group_id=consumer.group_id,
            msg=msg_mock,
        )
        mock_produce_for.assert_called_once_with(
            header_message=str(exc),
            header_detail=traceback.format_exc(),
        )

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_auto_offset_false(self, mock_consumer_client):
        class SomeConsumer(Consumer):
            config = {"enable.auto.offset.store": False}

        consumer = SomeConsumer()
        msg = Mock()

        consumer.commit_offset(msg)

        mock_consumer_client.return_value.store_offsets.assert_called_once_with(msg)

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_auto_offset_true(self, mock_consumer_client):
        class SomeConsumer(Consumer):
            config = {"enable.auto.offset.store": True}

        consumer = SomeConsumer()

        consumer.commit_offset(Mock())

        mock_consumer_client.return_value.store_offsets.assert_not_called()

    def test_settings_are_correctly_assigned(self):
        self.assertEqual(Consumer.polling_freq, settings.POLLING_FREQUENCY)
        self.assertEqual(Consumer.default_error_handler, settings.ERROR_HANDLER)

    @override_settings(
        **{
            SETTINGS_KEY: {
                "CLIENT_ID": "client.id-initial-definition",
                "GLOBAL_CONFIG": {
                    "client.id": "client-id-overridden-by-global-config",
                    "bootstrap.servers": "defined-in-global-config",
                },
                "CONSUMER_CONFIG": {
                    "bootstrap.servers": "bootstrap.servers-overridden-by-consumer",
                    "group.id": "group.id-defined-by-consumer-config",
                },
                "ERROR_HANDLER": "django_kafka.error_handlers.ClientErrorHandler",
            },
        },
    )
    @patch("django_kafka.consumer.ConfluentConsumer")
    @patch("django_kafka.error_handlers.ClientErrorHandler")
    def test_config_merge_override(self, mock_error_handler, mock_consumer_client):
        """
        1. CLIENT_ID is added to the consumers config
        2. GLOBAL_CONFIG overrides CLIENT_ID (client.id) if it contains one
        3. CONSUMER_CONFIG merges with GLOBAL_CONFIG and overrides keys if any
        4. Consumer.config is merged with CONSUMER_CONFIG and overrides keys if any
        """

        class SomeConsumer(Consumer):
            config = {
                "group.id": "group.id.overridden-by-consumer-class",
                "enable.auto.offset.store": True,
            }

        # Consumer.config is properly merged
        consumer = SomeConsumer()

        self.assertDictEqual(
            SomeConsumer.build_config(),
            {
                "client.id": "client-id-overridden-by-global-config",
                "bootstrap.servers": "bootstrap.servers-overridden-by-consumer",
                "group.id": "group.id.overridden-by-consumer-class",
                "enable.auto.offset.store": True,
                "logger": SomeConsumer.default_logger,
                "error_cb": mock_error_handler(),
            },
        )
        self.assertDictEqual(
            SomeConsumer.build_config(),
            consumer.config,
        )



import datetime


def days_from_epoch_to_date(days: int, default=None) -> datetime.date | None:
    """
    DATE    INT32   io.debezium.time.Date
                    Represents the number of days since the epoch.
    [Temporal types](https://debezium.io/documentation/reference/stable/connectors/postgresql.html#postgresql-temporal-types)
    """
    if days is None or days == "":
        return default
    return datetime.date(1970, 1, 1) + datetime.timedelta(days)



from unittest.mock import patch, call

from django.test import SimpleTestCase

from django_kafka.exceptions import DjangoKafkaError
from django_kafka.connect.client import KafkaConnectSession, KafkaConnectClient


class KafkaConnectSessionTestCase(SimpleTestCase):
    @patch("django_kafka.connect.client.KafkaConnectSession.mount_retry_adapter")
    def test_session_initializes_retry_not_provided(self, mock_mount_retry_adapter):
        KafkaConnectSession("http://localhost")
        mock_mount_retry_adapter.assert_not_called()

    @patch("django_kafka.connect.client.KafkaConnectSession.mount_retry_adapter")
    def test_session_initializes_retry(self, mock_mount_retry_adapter):
        retry_kwargs = {
            "connect": 1,
            "read": 1,
            "status": 1,
            "backoff_factor": 0.1,
            "status_forcelist": [500],
        }

        KafkaConnectSession("http://localhost", retry=retry_kwargs)

        mock_mount_retry_adapter.assert_called_once_with(retry_kwargs)

    @patch("django_kafka.connect.client.super")
    def test_session_request(self, mock_super):
        host = "http://localhost"
        request_path = "/path"
        
        session = KafkaConnectSession(host, timeout=10)

        # should use timeout value from the init
        session.request("GET", request_path)
        # should use the value of the provided timeout kwarg
        session.request("GET", request_path, timeout=20)

        # url was correctly constructed and right timeout was used
        mock_super().request.assert_has_calls([
            call("GET", f"{host}{request_path}", timeout=10),
            call("GET", f"{host}{request_path}", timeout=20),
        ], any_order=False)


class KafkaConnectClientTestCase(SimpleTestCase):
    @patch("django_kafka.connect.client.KafkaConnectSession", spec=True)
    def test_init_request_session(self, mock_session):
        init_args = "http://localhost", None, None, 10
        client = KafkaConnectClient(*init_args)
        mock_session.assert_called_once_with(*init_args)
        self.assertIsInstance(client._requests, KafkaConnectSession)

    @patch("django_kafka.connect.client.KafkaConnectSession")
    def test_validate_connector_class_not_in_config(self, mock_session):
        client = KafkaConnectClient("http://localhost")

        with self.assertRaises(DjangoKafkaError):
            client.validate({})

        mock_session.return_value.put.assert_not_called()

    @patch("django_kafka.connect.client.KafkaConnectSession")
    def test_validate_parse_connector_class(self, mock_session):
        client = KafkaConnectClient("http://localhost")
        config = {"connector.class": "io.debezium.connector.postgresql.PostgresConnector"}
        connector_class_name = "PostgresConnector"

        result = client.validate(config)

        mock_session.return_value.put.assert_called_once_with(
            f"/connector-plugins/{connector_class_name}/config/validate",
            json=config,
        )
        self.assertEqual(result, mock_session.return_value.put.return_value)

    @patch("django_kafka.connect.client.KafkaConnectSession")
    def test_update_or_create(self, mock_session):
        client = KafkaConnectClient("http://localhost")
        connector_name = "connector_name"
        config = {"key": "value"}

        result = client.update_or_create(connector_name, config)

        mock_session.return_value.put.assert_called_once_with(
            f"/connectors/{connector_name}/config",
            json=config,
        )
        self.assertEqual(result, mock_session.return_value.put.return_value)

    @patch("django_kafka.connect.client.KafkaConnectSession")
    def test_delete(self, mock_session):
        client = KafkaConnectClient("http://localhost")
        connector_name = "connector_name"

        result = client.delete(connector_name)

        mock_session.return_value.delete.assert_called_once_with(f"/connectors/{connector_name}")
        self.assertEqual(result, mock_session.return_value.delete.return_value)

    @patch("django_kafka.connect.client.KafkaConnectSession")
    def test_connector_status(self, mock_session):
        client = KafkaConnectClient("http://localhost")
        connector_name = "connector_name"

        result = client.connector_status(connector_name)

        mock_session.return_value.get.assert_called_once_with(f"/connectors/{connector_name}/status")
        self.assertEqual(result, mock_session.return_value.get.return_value)



from abc import ABC

from confluent_kafka.schema_registry.avro import AvroDeserializer, AvroSerializer

from django_kafka import kafka
from django_kafka.topic import Topic, TopicConsumer, TopicProducer


class AvroTopicProducer(TopicProducer, ABC):
    """
    Defining `value_schema` is required (`key_schema` is required when using keys).
    It gets submitted to the Schema Registry.

    Multiple schemas and one Topic:
        `.produce(...)` takes `{key|value}_serializer_kwargs` kw argument.
        `AvroSerializer` then gets initialized with the provided kwargs.
        When producing you can tell which schema to use for your message:
        ```python
        schema = {
            "type": "record",
            "name": "ValueTest",
            "fields": [
                {"name": "value", "type": "string"},
            ]
        }
        topic.produce({"value": 1}, value_serializer_kwargs={"schema_str": json.dumps(schema)})
        ```

    [Cofluent AvroSerializer Config](https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#avroserializer)
    [Avro schema definition](https://avro.apache.org/docs/1.11.1/specification/)
    """  # noqa: E501

    key_schema: str
    value_schema: str
    schema_config: dict

    def get_key_serializer(self, **kwargs):
        kwargs.setdefault("schema_str", getattr(self, "key_schema", None))
        kwargs.setdefault("conf", getattr(self, "schema_config", None))

        return AvroSerializer(kafka.schema_client, **kwargs)

    def get_value_serializer(self, **kwargs):
        kwargs.setdefault("schema_str", getattr(self, "value_schema", None))
        kwargs.setdefault("conf", getattr(self, "schema_config", None))

        return AvroSerializer(kafka.schema_client, **kwargs)


class AvroTopicConsumer(TopicConsumer, ABC):
    """
    Defining schemas is not necessary as it gets retrieved automatically from the Schema Registry.
    """  # noqa: E501

    def get_key_deserializer(self, **kwargs):
        return AvroDeserializer(kafka.schema_client, **kwargs)

    def get_value_deserializer(self, **kwargs):
        return AvroDeserializer(kafka.schema_client, **kwargs)


class AvroTopic(AvroTopicConsumer, AvroTopicProducer, Topic, ABC):
    """Combined AvroTopic class for when topic consume and produce names are the same"""



from abc import ABC
from typing import Optional, Type

from django.db.models import Model

from django_kafka.exceptions import DjangoKafkaError
from django_kafka.topic.model import ModelTopicConsumer


class DbzModelTopicConsumer(ModelTopicConsumer, ABC):
    """Syncs a debezium source connector topic directly in to Django model instances"""

    reroute_key_field_name = "__dbz__physicalTableIdentifier"  # see: https://debezium.io/documentation/reference/stable/transformations/topic-routing.html#by-logical-table-router-key-field-name
    reroute_model_map: Optional[dict[str, Type[Model]]] = None

    def get_model(self, key, value) -> Type[Model]:
        if self.reroute_key_field_name in key:
            if not self.reroute_model_map:
                raise DjangoKafkaError(
                    f"To obtain the correct model, reroute_model_map must be set when "
                    f"`{self.reroute_key_field_name}` is present in the message key",
                )
            table = key[self.reroute_key_field_name]
            if table not in self.reroute_model_map:
                raise DjangoKafkaError(f"Unrecognised rerouted topic `{table}`")
            return self.reroute_model_map[table]
        return super().get_model(key, value)

    def is_deletion(self, model, key, value) -> bool:
        if value is None:
            return True

        deleted = value.pop("__deleted", None)
        if isinstance(deleted, bool):
            return deleted
        if isinstance(deleted, str):
            return deleted.lower() == "true"
        return False

    def get_lookup_kwargs(self, model, key, value) -> dict:
        pk_field = model._meta.pk.name
        return {pk_field: key[pk_field]}



from unittest.mock import patch

from django.test import TestCase, override_settings

from django_kafka.conf import SETTINGS_KEY
from django_kafka.topic.avro import AvroTopicConsumer, AvroTopicProducer


class ATopicProducer(AvroTopicProducer):
    name = "some_topic_with_avro_serialization"


@override_settings(
    **{SETTINGS_KEY: {"SCHEMA_REGISTRY": {"url": "http://schema-registy"}}},
)
@patch("django_kafka.topic.kafka.schema_client")
class AvroTopicTestCase(TestCase):
    def setUp(self):
        self.topic_producer = ATopicProducer()

    @patch("django_kafka.topic.avro.AvroSerializer")
    def test_get_key_serializer(
        self,
        mock_avro_serializer,
        mock_kafka_schema_client,
    ):
        kwargs = {
            "schema_str": "<some schema>",
            "conf": {},
        }
        key_serializer = self.topic_producer.get_key_serializer(**kwargs)

        # returns AvroSerializer instance
        self.assertEqual(key_serializer, mock_avro_serializer.return_value)
        # instance was initialized with right arguments
        mock_avro_serializer.assert_called_once_with(
            mock_kafka_schema_client,
            **kwargs,
        )

    @patch("django_kafka.topic.avro.AvroSerializer")
    def test_get_value_serializer(
        self,
        mock_avro_serializer,
        mock_kafka_schema_client,
    ):
        kwargs = {
            "schema_str": "<some schema>",
            "conf": {},
        }
        value_serializer = self.topic_producer.get_value_serializer(**kwargs)

        # returns AvroSerializer instance
        self.assertEqual(value_serializer, mock_avro_serializer.return_value)
        # instance was initialized with right arguments
        mock_avro_serializer.assert_called_once_with(
            mock_kafka_schema_client,
            **kwargs,
        )


class ATopicConsumer(AvroTopicConsumer):
    name = "some_topic_with_avro_deserialization"

    def consume(self, msg):
        pass


@override_settings(
    **{SETTINGS_KEY: {"SCHEMA_REGISTRY": {"url": "http://schema-registy"}}},
)
@patch("django_kafka.topic.kafka.schema_client")
class AvroTopicConsumerTestCase(TestCase):
    def setUp(self):
        self.topic_consumer = ATopicConsumer()

    @patch("django_kafka.topic.avro.AvroDeserializer")
    def test_get_key_deserializer(
        self,
        mock_avro_deserializer,
        mock_kafka_schema_client,
    ):
        kwargs = {}
        key_deserializer = self.topic_consumer.get_key_deserializer(**kwargs)

        # returns mock_AvroDeserializer instance
        self.assertEqual(key_deserializer, mock_avro_deserializer.return_value)
        # instance was initialized with right arguments
        mock_avro_deserializer.assert_called_once_with(
            mock_kafka_schema_client,
            **kwargs,
        )

    @patch("django_kafka.topic.avro.AvroDeserializer")
    def test_get_value_deserializer(
        self,
        mock_avro_deserializer,
        mock_kafka_schema_client,
    ):
        kwargs = {}
        value_deserializer = self.topic_consumer.get_value_deserializer(**kwargs)

        # returns mock_AvroDeserializer instance
        self.assertEqual(value_deserializer, mock_avro_deserializer.return_value)
        # instance was initialized with right arguments
        mock_avro_deserializer.assert_called_once_with(
            mock_kafka_schema_client,
            **kwargs,
        )



from unittest.mock import patch

from django_kafka.models import KafkaSkipModel
from django_kafka.tests.models import AbstractModelTestCase


class KafkaSkipModelTestCase(AbstractModelTestCase):
    abstract_model = KafkaSkipModel
    model: type[KafkaSkipModel]

    def test_save__direct_instance_respects_set_kafka_skip(self):
        """test `save` on directly created instances will not ignore set kafka_skip"""
        instance = self.model(kafka_skip=True)
        instance_with_pk = self.model(pk=2, kafka_skip=True)

        instance.save()
        instance_with_pk.save()

        self.assertTrue(instance.kafka_skip)
        self.assertTrue(instance_with_pk.kafka_skip)

    def test_save__second_save_resets_kafka_skip(self):
        """test saving a second time will reset kafka_skip to False"""
        instance = self.model(kafka_skip=True)

        instance.save()
        instance.save()

        self.assertFalse(instance.kafka_skip)

    def test_save__adds_kafka_skip_to_update_fields_when_reset(self):
        """test update_fields has kafka_skip added when kafka_skip is reset"""
        instance = self.model(kafka_skip=True)
        instance.save()

        with patch("django.db.models.Model.save") as mock_save:
            instance.save(update_fields=["field"])

        mock_save.assert_called_once_with(
            force_insert=False,
            force_update=False,
            using=None,
            update_fields=["kafka_skip", "field"],
        )

    def test__saving_retrieved_db_instance__resets_unset_kafka_skip(self):
        """test saving a db-retrieved instance will reset kafka_skip if not set"""
        instance = self.model(kafka_skip=True)
        instance.save()

        retrieved = self.model.objects.get(pk=instance.pk)
        retrieved.save()

        self.assertFalse(retrieved.kafka_skip)

    def test__saving_retrieved_db_instance__respects_set_kafka_skip(self):
        """test saving a db-retrieved instance will not ignore newly set kafka skip"""
        instance = self.model(kafka_skip=False)
        instance.save()

        retrieved = self.model.objects.get(pk=instance.pk)
        retrieved.kafka_skip = True
        retrieved.save()

        self.assertTrue(retrieved.kafka_skip)

    def test__refresh_from_db__makes_kafka_skip_be_reset(self):
        """test refreshing from db will ignore newly set kafka_skip"""
        instance = self.model(kafka_skip=False)
        instance.save()

        instance.kafka_skip = True
        instance.refresh_from_db()
        instance.save()

        self.assertFalse(instance.kafka_skip)

    def test_queryset_create__respects_kafka_skip(self):
        """test kafka_skip=True is maintained for QuerySet.create"""
        instance = self.model.objects.create(kafka_skip=True)
        instance_with_pk = self.model.objects.create(pk=2, kafka_skip=True)

        self.assertTrue(instance.kafka_skip)
        self.assertTrue(instance_with_pk.kafka_skip)

    def test_queryset_get_or_create__respects_kafka_skip(self):
        """test kafka_skip=True is maintained for QuerySet.get_or_create"""
        instance, _ = self.model.objects.get_or_create(defaults={"kafka_skip": True})
        instance_with_pk, _ = self.model.objects.get_or_create(
            pk=2,
            defaults={"kafka_skip": True},
        )

        self.assertTrue(instance.kafka_skip)
        self.assertTrue(instance_with_pk.kafka_skip)

    def test_queryset_update_or_create_with_kafka_skip(self):
        """test kafka_skip=True is maintained for QuerySet.update_or_create"""
        instance = self.model.objects.create(kafka_skip=False)

        existing_instance, _ = self.model.objects.update_or_create(
            pk=instance.id,
            defaults={"kafka_skip": True},
        )
        new_instance, _ = self.model.objects.update_or_create(
            pk=2,
            defaults={"kafka_skip": True},
        )

        self.assertTrue(existing_instance.kafka_skip)
        self.assertTrue(new_instance.kafka_skip)

    @patch("django.db.models.QuerySet.update")
    def test_queryset_update_sets_kafka_skip(self, mock_update):
        """
        `update` method should automatically set `kafka_skip=False`
          if `kafka_skip` is not provided in kwargs.
        """
        # update_kwargs does not contain kafka_skip
        update_kwargs = {"some_field": "some_value"}

        self.model.objects.update(**update_kwargs)

        # kafka_skip=False was added to the update_kwargs
        mock_update.assert_called_once_with(
            **{"kafka_skip": False, **update_kwargs},
        )

    @patch("django.db.models.QuerySet.update")
    def test_queryset_update_does_not_override_kafka_skip(self, mock_update):
        """
        `update` method should not change `kafka_skip`
          if `kafka_skip` is provided in kwargs
        """
        # update_kwargs contains kafka_skip
        update_kwargs = {"kafka_skip": True, "some_field": "some_value"}

        self.model.objects.update(**update_kwargs)

        # kafka_skip is not changed, update_kwargs are not changed
        mock_update.assert_called_once_with(**update_kwargs)



import datetime
import traceback
from typing import Type
from unittest.mock import Mock, patch

from confluent_kafka import TopicPartition
from django.test import TestCase, override_settings
from django.utils import timezone

from django_kafka.conf import SETTINGS_KEY
from django_kafka.consumer import Consumer, Topics
from django_kafka.retry.consumer import RetryConsumer, RetryTopics
from django_kafka.retry.header import RetryHeader
from django_kafka.retry.settings import RetrySettings
from django_kafka.topic import TopicConsumer


class RetryConsumerTestCase(TestCase):
    def _get_topic_consumer(self):
        class SomeTopicConsumer(TopicConsumer):
            name = "normal_topic"

        return SomeTopicConsumer()

    def _get_retryable_topic_consumer(self):
        class RetryableTopicConsumer(TopicConsumer):
            name = "retry_topic"

        retry = RetrySettings(max_retries=5, delay=60)
        retry(RetryableTopicConsumer)

        return RetryableTopicConsumer()

    def _get_retryable_consumer_cls(self, group_id="group_id") -> Type[Consumer]:
        class SomeConsumer(Consumer):
            topics = Topics(
                self._get_topic_consumer(),
                self._get_retryable_topic_consumer(),
            )
            config = {"group.id": group_id}

        return SomeConsumer

    def _get_retry_consumer(self, consumer_group_id="group_id") -> RetryConsumer:
        return RetryConsumer.build(
            self._get_retryable_consumer_cls(group_id=consumer_group_id),
        )()

    @override_settings(
        **{
            SETTINGS_KEY: {
                "RETRY_CONSUMER_CONFIG": {
                    "bootstrap.servers": "bootstrap.defined-by-retry-consumer-config",
                    "group.id": "group.id-defined-by-retry-consumer-config",
                    "topic.metadata.refresh.interval.ms": 10000,
                },
            },
        },
    )
    @patch("django_kafka.consumer.ConfluentConsumer")
    @patch("django_kafka.retry.consumer.Consumer.build_config")
    def test_config_merge_override(
        self,
        mock_consumer_build_config,
        mock_consumer_client,
    ):
        """
        1. Consumer.build_config() is added to the consumers config
        2. RETRY_CONSUMER_CONFIG is merged next and overrides keys if any
        3. RetryConsumer.config is merged next and overrides keys if any
        """

        mock_consumer_build_config.return_value = {
            "bootstrap.servers": "bootstrap.defined-by-consumer-cls",
            "group.id": "group.id.set-by-consumer-cls",
            "enable.auto.offset.store": False,
        }

        class SomeRetryConsumer(RetryConsumer):
            config = {
                "group.id": "group.id.overridden-by-retry-consumer-class",
            }

        retry_consumer = SomeRetryConsumer()

        self.assertDictEqual(
            SomeRetryConsumer.build_config(),
            {
                "bootstrap.servers": "bootstrap.defined-by-retry-consumer-config",
                "group.id": "group.id.overridden-by-retry-consumer-class",
                "topic.metadata.refresh.interval.ms": 10000,
                "enable.auto.offset.store": False,
            },
        )
        self.assertDictEqual(
            SomeRetryConsumer.build_config(),
            retry_consumer.config,
        )

    def test_build(self):
        consumer_cls = self._get_retryable_consumer_cls()
        retry_consumer_cls = RetryConsumer.build(consumer_cls)

        self.assertTrue(issubclass(retry_consumer_cls, RetryConsumer))

        self.assertEqual(
            retry_consumer_cls.config["group.id"],
            f"{consumer_cls.build_config()['group.id']}.retry",
        )
        self.assertIsInstance(retry_consumer_cls.topics, RetryTopics)
        self.assertCountEqual(
            [t for t in consumer_cls.topics if t.retry_settings],
            [t.topic_consumer for t in retry_consumer_cls.topics],
        )

    def test_build__no_retry_topics(self):
        class TestConsumer(Consumer):
            topics = Topics()

        self.assertIsNone(RetryConsumer.build(TestConsumer))

    def test_retry_msg(self):
        mock_retry_topic_consumer = Mock()
        mock_producer_for = mock_retry_topic_consumer.producer_for
        mock_retry = mock_producer_for.return_value.retry
        msg_mock = Mock()

        retry_consumer = self._get_retry_consumer()
        retry_consumer.get_topic_consumer = Mock(return_value=mock_retry_topic_consumer)
        exc = ValueError()

        retried = retry_consumer.retry_msg(msg_mock, exc)

        mock_producer_for.assert_called_once_with(msg_mock)
        mock_retry.assert_called_once_with(exc)
        self.assertEqual(retried, mock_retry.return_value)

    @patch("django_kafka.retry.consumer.DeadLetterTopicProducer")
    def test_dead_letter_msg(self, mock_dlt_topic_producer_cls):
        mock_retry_topic_consumer = Mock()
        mock_produce_for = mock_dlt_topic_producer_cls.return_value.produce_for
        msg_mock = Mock()

        retry_consumer = self._get_retry_consumer()
        retry_consumer.get_topic_consumer = Mock(return_value=mock_retry_topic_consumer)
        exc = ValueError()

        retry_consumer.dead_letter_msg(msg_mock, exc)

        mock_dlt_topic_producer_cls.assert_called_once_with(
            group_id=mock_retry_topic_consumer.group_id,
            msg=msg_mock,
        )
        mock_produce_for.assert_called_once_with(
            header_message=str(exc),
            header_detail=traceback.format_exc(),
        )

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_pause_partition(self, mock_confluent_consumer):
        retry_consumer = self._get_retry_consumer()
        mock_msg = Mock(
            **{
                "topic.return_value": "msg_topic",
                "partition.return_value": 0,
                "offset.return_value": 0,
            },
        )
        partition = TopicPartition(
            mock_msg.topic(),
            mock_msg.partition(),
            mock_msg.offset(),
        )
        retry_time = timezone.now()

        retry_consumer.pause_partition(mock_msg, retry_time)

        mock_confluent_consumer.return_value.seek.assert_called_once_with(partition)
        mock_confluent_consumer.return_value.pause.assert_called_once_with([partition])

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_resume_partition__before_retry_time(self, mock_confluent_consumer):
        retry_consumer = self._get_retry_consumer()
        mock_msg = Mock(
            **{
                "topic.return_value": "msg_topic",
                "partition.return_value": 0,
                "offset.return_value": 0,
            },
        )
        retry_time = timezone.now() + datetime.timedelta(minutes=1)

        retry_consumer.pause_partition(mock_msg, retry_time)
        retry_consumer.resume_ready_partitions()

        mock_confluent_consumer.return_value.resume.assert_not_called()

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_resume_ready_partitions__after_retry_time(self, mock_confluent_consumer):
        retry_consumer = self._get_retry_consumer()
        mock_msg = Mock(
            **{
                "topic.return_value": "msg_topic",
                "partition.return_value": 0,
                "offset.return_value": 0,
            },
        )
        partition = TopicPartition(
            mock_msg.topic(),
            mock_msg.partition(),
            mock_msg.offset(),
        )
        retry_time = timezone.now() - datetime.timedelta(minutes=1)

        retry_consumer.pause_partition(mock_msg, retry_time)
        retry_consumer.resume_ready_partitions()

        mock_confluent_consumer.return_value.resume.assert_called_once_with([partition])

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_poll(self, mock_confluent_consumer):
        """tests poll resumes partitions"""
        retry_consumer = self._get_retry_consumer()
        retry_consumer.resume_ready_partitions = Mock()
        mock_msg = Mock()
        mock_confluent_consumer.return_value.poll.return_value = mock_msg

        msg = retry_consumer.poll()

        self.assertEqual(msg, mock_msg)
        retry_consumer.resume_ready_partitions.assert_called_once()  # always called

    @patch("django_kafka.consumer.Consumer.process_message")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message__before_retry_time(
        self,
        mock_confluent_consumer,
        mock_consumer_process_message,
    ):
        retry_consumer = self._get_retry_consumer()
        retry_consumer.pause_partition = Mock()
        retry_time = timezone.now() + datetime.timedelta(minutes=1)
        mock_msg = Mock(
            **{
                "error.return_value": None,
                "headers.return_value": [
                    (RetryHeader.TIMESTAMP, str(retry_time.timestamp())),
                ],
            },
        )

        retry_consumer.process_message(mock_msg)
        retry_consumer.pause_partition.assert_called_once_with(mock_msg, retry_time)
        mock_consumer_process_message.process_message.assert_not_called()

    @patch("django_kafka.consumer.Consumer.process_message")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message__after_retry_time(
        self,
        mock_confluent_consumer,
        mock_consumer_process_message,
    ):
        retry_consumer = self._get_retry_consumer()
        retry_consumer.pause_partition = Mock()
        retry_time = timezone.now() - datetime.timedelta(minutes=1)
        mock_msg = Mock(
            **{
                "error.return_value": None,
                "headers.return_value": [
                    (RetryHeader.TIMESTAMP, str(retry_time.timestamp())),
                ],
            },
        )

        retry_consumer.process_message(mock_msg)

        retry_consumer.pause_partition.assert_not_called()
        mock_consumer_process_message.assert_called_once_with(mock_msg)



from unittest.mock import patch

from django.test import override_settings, SimpleTestCase

from django_kafka.conf import DEFAULTS, SETTINGS_KEY, settings


class SettingsTestCase(SimpleTestCase):
    settings_keys = (
        "CLIENT_ID",
        "ERROR_HANDLER",
        "GLOBAL_CONFIG",
        "PRODUCER_CONFIG",
        "CONSUMER_CONFIG",
        "RETRY_CONSUMER_CONFIG",
        "RETRY_TOPIC_SUFFIX",
        "DEAD_LETTER_TOPIC_SUFFIX",
        "POLLING_FREQUENCY",
        "SCHEMA_REGISTRY",
        "CONNECT_HOST",
        "CONNECT_AUTH",
        "CONNECT_RETRY",
        "CONNECT_REQUESTS_TIMEOUT",
        "CONNECTOR_NAME_PREFIX",
    )

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_defaults(self, mock_consumer_client):
        # make sure defaults are assigned
        for key in self.settings_keys:
            self.assertEqual(getattr(settings, key), DEFAULTS[key])

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_user_settings(self, mock_consumer_client):
        # make sure settings defined by user pulled up
        user_settings = {
            "CLIENT_ID": "client-id",
            "ERROR_HANDLER": "error.handler.class",
            "GLOBAL_CONFIG": {"bootstrap.servers": "kafka1"},
            "PRODUCER_CONFIG": {
                "enable.idempotence": True,
            },
            "CONSUMER_CONFIG": {
                "group.id": "group-1",
            },
            "RETRY_CONSUMER_CONFIG": {
                "topic.metadata.refresh.interval.ms": 5000,
            },
            "RETRY_TOPIC_SUFFIX": "retry-extra",
            "DEAD_LETTER_TOPIC_SUFFIX": "dlt-extra",
            "POLLING_FREQUENCY": 0.5,
            "SCHEMA_REGISTRY": {
                "url": "https://schema-registry",
            },
            "CONNECT_HOST": "http://kafka-connect",
            "CONNECT_AUTH": ("user", "pass"),
            "CONNECT_RETRY": {
                "connect": 10,
                "read": 10,
                "status": 10,
                "backoff_factor": 0.1,
                "status_forcelist": [502, 503, 504],
            },
            "CONNECT_REQUESTS_TIMEOUT": 60,
            "CONNECTOR_NAME_PREFIX": "project_name"
        }
        with override_settings(**{SETTINGS_KEY: user_settings}):
            for key in self.settings_keys:
                self.assertEqual(getattr(settings, key), user_settings[key])



from unittest import mock

from django.test import TestCase, override_settings

from django_kafka.conf import SETTINGS_KEY
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.retry.header import RetryHeader
from django_kafka.retry.settings import RetrySettings
from django_kafka.retry.topic import (
    RetryTopicConsumer,
    RetryTopicProducer,
)
from django_kafka.topic import TopicConsumer


@override_settings(**{SETTINGS_KEY: {"RETRY_TOPIC_SUFFIX": "test-retry"}})
class RetryTopicProducerTestCase(TestCase):
    def test__get_next_attempt(self):
        self.assertEqual(RetryTopicProducer.get_next_attempt("topic"), 1)
        self.assertEqual(
            RetryTopicProducer.get_next_attempt("group.id.topic.fake.5"),
            1,
        )
        self.assertEqual(
            RetryTopicProducer.get_next_attempt("group.id.topic.test-retry"),
            1,
        )
        self.assertEqual(
            RetryTopicProducer.get_next_attempt("group.id.topic.test-retry.0"),
            1,
        )
        self.assertEqual(
            RetryTopicProducer.get_next_attempt("group.id.topic.test-retry.2"),
            3,
        )
        self.assertEqual(
            RetryTopicProducer.get_next_attempt("group.id.topic.test-retry.10"),
            11,
        )

    def test_init(self):
        retry_settings = RetrySettings(max_retries=5, delay=60)
        mock_msg_topic_consumer = mock.Mock(**{"topic.return_value": "topic.name"})

        rt_producer = RetryTopicProducer(
            group_id="group.id",
            retry_settings=retry_settings,
            msg=mock_msg_topic_consumer,
        )

        self.assertEqual(rt_producer.group_id, "group.id")
        self.assertEqual(rt_producer.settings, retry_settings)
        self.assertEqual(rt_producer.msg, mock_msg_topic_consumer)
        self.assertEqual(rt_producer.attempt, 1)

    def test_name(self):
        retry_settings = RetrySettings(max_retries=5, delay=60)
        mock_msg_topic_consumer = mock.Mock(**{"topic.return_value": "topic.name"})
        mock_msg_rt_producer = mock.Mock(
            **{
                "topic.return_value": "group.id.topic.name.test-retry.1",
            },
        )

        rt_producer_1 = RetryTopicProducer(
            group_id="group.id",
            retry_settings=retry_settings,
            msg=mock_msg_topic_consumer,
        )

        rt_producer_2 = RetryTopicProducer(
            group_id="group.id",
            retry_settings=retry_settings,
            msg=mock_msg_rt_producer,
        )

        self.assertEqual(rt_producer_1.name, "group.id.topic.name.test-retry.1")
        self.assertEqual(rt_producer_2.name, "group.id.topic.name.test-retry.2")

    @override_settings(**{SETTINGS_KEY: {"RETRY_TOPIC_SUFFIX": "test-retry"}})
    def test_name__uses_settings(self):
        retry_settings = RetrySettings(max_retries=5, delay=60)
        mock_msg_topic_consumer = mock.Mock(**{"topic.return_value": "topic.name"})

        rt_producer = RetryTopicProducer(
            group_id="group.id",
            retry_settings=retry_settings,
            msg=mock_msg_topic_consumer,
        )

        self.assertEqual(rt_producer.name, "group.id.topic.name.test-retry.1")

    @mock.patch("django_kafka.retry.settings.RetrySettings.get_retry_timestamp")
    def test_retry__first_retry(self, mock_get_retry_timestamp: mock.Mock):
        mock_msg = mock.Mock(**{"topic.return_value": "msg_topic"})
        retry_settings = RetrySettings(max_retries=5, delay=60)
        rt_producer = RetryTopicProducer(
            group_id="group.id",
            retry_settings=retry_settings,
            msg=mock_msg,
        )
        rt_producer.produce = mock.Mock()

        retried = rt_producer.retry(exc=ValueError("error message"))

        self.assertTrue(retried)
        rt_producer.produce.assert_called_with(
            key=mock_msg.key(),
            value=mock_msg.value(),
            headers=[
                (RetryHeader.MESSAGE, "error message"),
                (RetryHeader.TIMESTAMP, mock_get_retry_timestamp.return_value),
            ],
        )
        mock_get_retry_timestamp.assert_called_once_with(1)

    @mock.patch("django_kafka.retry.settings.RetrySettings.get_retry_timestamp")
    def test_retry__last_retry(self, mock_get_retry_timestamp):
        mock_msg = mock.Mock(
            **{"topic.return_value": "group.id.msg_topic.test-retry.4"},
        )
        rt_producer = RetryTopicProducer(
            group_id="group.id",
            retry_settings=RetrySettings(max_retries=5, delay=60),
            msg=mock_msg,
        )
        rt_producer.produce = mock.Mock()

        retried = rt_producer.retry(exc=ValueError("error message"))

        self.assertTrue(retried)
        rt_producer.produce.assert_called_with(
            key=mock_msg.key(),
            value=mock_msg.value(),
            headers=[
                (RetryHeader.MESSAGE, "error message"),
                (RetryHeader.TIMESTAMP, mock_get_retry_timestamp.return_value),
            ],
        )
        mock_get_retry_timestamp.assert_called_once_with(5)

    def test_retry__no_more_retries(self):
        rt_producer = RetryTopicProducer(
            group_id="group.id",
            retry_settings=RetrySettings(max_retries=5, delay=60),
            msg=mock.Mock(**{"topic.return_value": "group.id.msg_topic.test-retry.5"}),
        )
        rt_producer.produce = mock.Mock()

        retried = rt_producer.retry(exc=ValueError())

        self.assertFalse(retried)
        rt_producer.produce.assert_not_called()

    def test_retry__no_retry_excluded_error(self):
        rt_producer = RetryTopicProducer(
            group_id="group.id",
            retry_settings=RetrySettings(max_retries=5, delay=60, exclude=[ValueError]),
            msg=mock.Mock(**{"topic.return_value": "msg_topic"}),
        )
        rt_producer.produce = mock.Mock()

        retried = rt_producer.retry(exc=ValueError())

        self.assertFalse(retried)
        rt_producer.produce.assert_not_called()


@override_settings(**{SETTINGS_KEY: {"RETRY_TOPIC_SUFFIX": "test-retry"}})
class RetryTopicConsumerTestCase(TestCase):
    def _get_retryable_topic_consumer(
        self,
        topic_name: str = "topic_name",
        **retry_kwargs,
    ):
        class SomeTopicConsumer(TopicConsumer):
            name = topic_name

        retry = RetrySettings(**{"max_retries": 5, "delay": 60, **retry_kwargs})
        topic_consumer_cls = retry(SomeTopicConsumer)
        return topic_consumer_cls()

    def test_init__raises_without_retry(self):
        class SomeTopicConsumer(TopicConsumer):
            name = "topic_name"

        with self.assertRaises(DjangoKafkaError):
            RetryTopicConsumer(group_id="group.id", topic_consumer=SomeTopicConsumer())

    def test_name(self):
        """assert name format and correct escaping of regex special characters"""
        topic_consumer = self._get_retryable_topic_consumer(topic_name="topic.name")
        rt_consumer = RetryTopicConsumer(
            group_id="group.id",
            topic_consumer=topic_consumer,
        )

        self.assertEqual(
            rt_consumer.name,
            r"^group\.id\.topic\.name\.test\-retry\.([0-9]+)$",
        )

    def test_name__regex(self):
        """tests regex topic names are correctly inserted in to retry topic regex"""
        topic_consumer = self._get_retryable_topic_consumer(
            topic_name="^topic_name|other_name$",
        )
        rt_consumer = RetryTopicConsumer(
            group_id="group.id",
            topic_consumer=topic_consumer,
        )

        self.assertEqual(
            rt_consumer.name,
            r"^group\.id\.(topic_name|other_name)\.test\-retry\.([0-9]+)$",
        )

    def test_consume(self):
        """tests RetryTopic uses the main topic consume method"""
        topic_consumer = self._get_retryable_topic_consumer()
        topic_consumer.consume = mock.Mock()
        mock_msg = mock.Mock()

        RetryTopicConsumer(group_id="group.id", topic_consumer=topic_consumer).consume(
            mock_msg,
        )

        topic_consumer.consume.assert_called_once_with(mock_msg)

    @mock.patch("django_kafka.retry.topic.RetryTopicProducer")
    def test_producer_for(self, mock_rt_producer):
        topic_consumer = self._get_retryable_topic_consumer()
        mock_msg = mock.Mock()

        rt_consumer = RetryTopicConsumer(
            group_id="group.id",
            topic_consumer=topic_consumer,
        )
        rt_producer = rt_consumer.producer_for(mock_msg)

        self.assertEqual(rt_producer, mock_rt_producer.return_value)
        mock_rt_producer.assert_called_once_with(
            group_id=rt_consumer.group_id,
            retry_settings=topic_consumer.retry_settings,
            msg=mock_msg,
        )



from contextlib import contextmanager
from io import StringIO
from unittest.mock import patch, Mock, DEFAULT, MagicMock, call

from django.core.management import call_command
from django.test import SimpleTestCase
from requests.exceptions import RetryError

from django_kafka.exceptions import DjangoKafkaError
from django_kafka.connect.connector import Connector, ConnectorStatus
from django_kafka.management.commands.kafka_connect import Command


@contextmanager
def patch_kafka_connectors(**attrs):
    attrs.setdefault('mark_for_removal', False)

    connector_instance = Mock(spec_set=Connector, **attrs)
    connector_class = Mock(spec_set=Connector, return_value=connector_instance)

    with patch("django_kafka.kafka.connectors", {'connector.fake.path': connector_class}):
        yield connector_instance


class KafkaConnectTestCase(SimpleTestCase):
    def setUp(self):
        self.connector_path = 'connector.fake.path'
        self.command_stdout = StringIO()
        self.command = Command(stdout=self.command_stdout)
        self.command.connectors = [self.connector_path]

    @patch.multiple(
        "django_kafka.management.commands.kafka_connect.Command",
        handle_validate=DEFAULT,
        handle_publish=DEFAULT,
        handle_status=DEFAULT,
    )
    def test_execute_order(self, handle_validate, handle_publish, handle_status):
        # [Tracking order of calls](https://docs.python.org/3/library/unittest.mock-examples.html#tracking-order-of-calls-and-less-verbose-call-assertions)
        manager = MagicMock()
        manager.attach_mock(handle_validate, "handle_validate")
        manager.attach_mock(handle_publish, "handle_publish")
        manager.attach_mock(handle_status, "handle_status")

        call_command("kafka_connect", validate=True, publish=True, check_status=True)

        handle_validate.assert_called_once_with()
        handle_publish.assert_called_once_with()
        handle_status.assert_called_once_with()

        manager.assert_has_calls([
            call.handle_validate(),
            call.handle_publish(),
            call.handle_status(),
        ], any_order=False)

    def test_handle_validate(self):
        with patch_kafka_connectors() as connector:
            self.command.handle_validate()

        connector.is_valid.assert_called_once_with(raise_exception=True)

        self.assertFalse(self.command.has_failures)

    def test_handle_validate_when_marked_for_removal(self):
        with patch_kafka_connectors(mark_for_removal=True) as connector:
            self.command.handle_validate()

        connector.is_valid.assert_not_called()
        self.assertFalse(self.command.has_failures)

    def test_handle_validate__exceptions(self):
        with patch_kafka_connectors(**{"is_valid.side_effect": DjangoKafkaError}) as connector:
            self.command.handle_validate()

        connector.is_valid.assert_called_once_with(raise_exception=True)
        self.assertTrue(self.command.has_failures)

    @patch("django_kafka.management.commands.kafka_connect.Command.handle_delete")
    @patch("django_kafka.management.commands.kafka_connect.Command.handle_submit")
    def test_handle_publish_marked_for_removal(self, mock_command_handle_submit, mock_command_handle_delete):
        with patch_kafka_connectors(mark_for_removal=True) as connector:
            self.command.handle_publish()

        mock_command_handle_submit.assert_not_called()
        mock_command_handle_delete.assert_called_once_with(connector)

    @patch("django_kafka.management.commands.kafka_connect.Command.handle_delete")
    @patch("django_kafka.management.commands.kafka_connect.Command.handle_submit")
    def test_handle_publish_not_marked_for_removal(self, mock_command_handle_submit, mock_command_handle_delete):
        with patch_kafka_connectors(mark_for_removal=False) as connector:
            self.command.handle_publish()

        mock_command_handle_submit.assert_called_once_with(connector)
        mock_command_handle_delete.assert_not_called()

    def test_handle_delete(self):
        with patch_kafka_connectors() as connector:
            self.command.handle_delete(connector)

        connector.delete.assert_called_once_with()
        self.assertFalse(self.command.has_failures)

    def test_handle_delete__django_kafka_error(self):
        with patch_kafka_connectors(**{"delete.side_effect": DjangoKafkaError}) as connector:
            self.command.handle_delete(connector)

        connector.delete.assert_called_once_with()
        self.assertTrue(self.command.has_failures)

    def test_handle_delete__retry_error_error(self):
        with patch_kafka_connectors(**{"delete.side_effect": RetryError}) as connector:
            self.command.handle_delete(connector)

        connector.delete.assert_called_once_with()
        self.assertTrue(self.command.has_failures)

    def test_handle_submit(self):
        with patch_kafka_connectors() as connector:
            self.command.handle_submit(connector)

        self.assertFalse(self.command.has_failures)
        connector.submit.assert_called_once_with()

    def test_handle_submit__django_kafka_error(self):
        with patch_kafka_connectors(**{"submit.side_effect": DjangoKafkaError}) as connector:
            self.command.handle_submit(connector)

        connector.submit.assert_called_once_with()
        self.assertTrue(self.command.has_failures)

    def test_handle_submit__retry_error_error(self):
        with patch_kafka_connectors(**{"submit.side_effect": RetryError}) as connector:
            self.command.handle_submit(connector)

        connector.submit.assert_called_once_with()
        self.assertTrue(self.command.has_failures)

    def test_handle_status__marked_for_removal(self):
        with patch_kafka_connectors(mark_for_removal=True) as connector:
            self.command.handle_status()

        connector.status.assert_not_called()

    def test_handle_status__running(self):
        with patch_kafka_connectors(**{"status.return_value": ConnectorStatus.RUNNING}) as connector:
            self.command.handle_status()

        connector.status.assert_called_once_with()
        self.assertFalse(self.command.has_failures)

    def test_handle_status__paused(self):
        with patch_kafka_connectors(**{"status.return_value": ConnectorStatus.PAUSED}) as connector:
            self.command.handle_status()

        connector.status.assert_called_once_with()
        self.assertTrue(self.command.has_failures)

    def test_handle_status__unassigned(self):
        with patch_kafka_connectors(**{"status.return_value": ConnectorStatus.UNASSIGNED}) as connector:
            self.command.handle_status()

        # status UNASSIGNED is retried 3 times
        self.assertEqual(connector.status.call_count, 3)

        self.assertTrue(self.command.has_failures)

    def test_handle_status__django_kafka_error(self):
        with patch_kafka_connectors(**{"status.side_effect": DjangoKafkaError}) as connector:
            self.command.handle_status()

        connector.status.assert_called_once_with()
        self.assertTrue(self.command.has_failures)

    def test_handle_status__retry_error(self):
        with patch_kafka_connectors(**{"status.side_effect": RetryError}) as connector:
            self.command.handle_status()

        connector.status.assert_called_once_with()
        self.assertTrue(self.command.has_failures)



from abc import ABC, abstractmethod
from enum import StrEnum

from django_kafka.conf import settings
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.connect.client import KafkaConnectClient

__all__ = [
    "Connector",
    "ConnectorStatus",
]


class ConnectorStatus(StrEnum):
    """
    https://docs.confluent.io/platform/current/connect/monitoring.html#connector-and-task-status
    UNASSIGNED: The connector/task has not yet been assigned to a worker.
    RUNNING: The connector/task is running.
    PAUSED: The connector/task has been administratively paused.
    FAILED: The connector/task has failed (usually by raising an exception, which is reported in the status output).
    """
    UNASSIGNED = "UNASSIGNED"
    RUNNING = "RUNNING"
    PAUSED = "PAUSED"


class Name:
    def __get__(self, instance, owner):
        if settings.CONNECTOR_NAME_PREFIX:
            return f"{settings.CONNECTOR_NAME_PREFIX}.{owner.__name__}"
        return owner.__name__


class Connector(ABC):
    name = Name()
    mark_for_removal = False

    @property
    @abstractmethod
    def config(self) -> dict:
        """Configurations for the connector."""

    def __init__(self):
        if not settings.CONNECT_HOST:
            raise DjangoKafkaError("Kafka `CONNECT_HOST` is not configured.")

        self.client = KafkaConnectClient(
            host=settings.CONNECT_HOST,
            auth=settings.CONNECT_AUTH,
            retry=settings.CONNECT_RETRY,
            timeout=settings.CONNECT_REQUESTS_TIMEOUT,
        )

    def delete(self) -> bool:
        response = self.client.delete(self.name)

        if response.status_code == 404:
            return False

        if not response.ok:
            raise DjangoKafkaError(response.text, context=response)

        return True

    def submit(self) -> dict:
        response = self.client.update_or_create(self.name, self.config)

        if not response.ok:
            raise DjangoKafkaError(response.text, context=response)

        return response.json()

    def is_valid(self, raise_exception=False) -> bool:
        response = self.client.validate(self.config)

        if raise_exception and not response.ok:
            raise DjangoKafkaError(response.text, context=response)

        return response.ok

    def status(self) -> ConnectorStatus:
        response = self.client.connector_status(self.name)

        if not response.ok:
            raise DjangoKafkaError(response.text, context=response)
        return response.json()["connector"]["state"]



from unittest import mock

from django.db.models import Model
from django.test import TestCase

from django_kafka.exceptions import DjangoKafkaError
from django_kafka.topic.debezium import DbzModelTopicConsumer


class TestDbzModelTopicConsumer(TestCase):
    def _get_model_topic_consumer(self) -> DbzModelTopicConsumer:
        class SomeModelTopicConsumer(DbzModelTopicConsumer):
            name = "name"
            model = Model

        return SomeModelTopicConsumer()

    def test_get_model__uses_reroute_model_map(self):
        mock_model = mock.Mock()
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.reroute_model_map = {"topic": mock_model}
        topic_consumer.reroute_key_field_name = "table_identifier_key"

        self.assertEqual(
            topic_consumer.get_model({"table_identifier_key": "topic"}, {}),
            mock_model,
        )

    def test_get_model__uses_direct_model(self):
        mock_model = mock.Mock()
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = mock_model
        topic_consumer.reroute_model_map = None

        self.assertEqual(topic_consumer.get_model({}, {}), mock_model)

    def test_get_model__raises_when_nothing_set(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = None
        topic_consumer.reroute_model_map = None

        with self.assertRaises(DjangoKafkaError):
            topic_consumer.get_model({}, {})

    def test_get_model__raises_when_reroute_key_present(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = Model
        topic_consumer.reroute_model_map = None
        topic_consumer.reroute_key_field_name = "table_identifier_key"

        with self.assertRaises(DjangoKafkaError):
            topic_consumer.get_model({"table_identifier_key": "table"}, {})

    def test_is_deletion(self):
        topic_consumer = self._get_model_topic_consumer()

        self.assertEqual(
            topic_consumer.is_deletion(Model, {}, {}),
            False,
        )
        self.assertEqual(
            topic_consumer.is_deletion(Model, {}, {"__deleted": "true"}),
            True,
        )
        self.assertEqual(
            topic_consumer.is_deletion(Model, {}, {"__deleted": "True"}),
            True,
        )
        self.assertEqual(
            topic_consumer.is_deletion(Model, {}, {"__deleted": "false"}),
            False,
        )
        self.assertEqual(
            topic_consumer.is_deletion(Model, {}, {"__deleted": "anything"}),
            False,
        )
        self.assertEqual(
            topic_consumer.is_deletion(Model, {}, {"__deleted": False}),
            False,
        )
        self.assertEqual(
            topic_consumer.is_deletion(
                Model,
                {},
                {"name": "name", "__deleted": True},
            ),
            True,
        )
        self.assertEqual(topic_consumer.is_deletion(Model, {}, None), True)

    def test_get_lookup_kwargs(self):
        topic_consumer = self._get_model_topic_consumer()
        mock_model = mock.Mock(**{"_meta.pk.name": "identity_key"})

        self.assertEqual(
            topic_consumer.get_lookup_kwargs(mock_model, {"identity_key": 1}, {}),
            {"identity_key": 1},
        )

    def test_get_lookup_kwargs__raises_when_pk_not_present(self):
        topic_consumer = self._get_model_topic_consumer()
        mock_model = mock.Mock(**{"_meta.pk.name": "identity_key"})

        with self.assertRaises(KeyError):
            topic_consumer.get_lookup_kwargs(mock_model, {"something": 1}, {})



import socket

from django.conf import settings as django_settings

SETTINGS_KEY = "DJANGO_KAFKA"
DEFAULTS = {
    "CLIENT_ID": f"{socket.gethostname()}-python",
    "ERROR_HANDLER": "django_kafka.error_handlers.ClientErrorHandler",
    "GLOBAL_CONFIG": {},
    "PRODUCER_CONFIG": {},
    "CONSUMER_CONFIG": {},
    "RETRY_CONSUMER_CONFIG": {
        "auto.offset.reset": "earliest",
        "enable.auto.offset.store": False,
        "topic.metadata.refresh.interval.ms": 10000,
    },
    "RETRY_TOPIC_SUFFIX": "retry",
    "DEAD_LETTER_TOPIC_SUFFIX": "dlt",
    "POLLING_FREQUENCY": 1,  # seconds
    "SCHEMA_REGISTRY": {},
    # Rest API of the kafka-connect instance
    "CONNECT_HOST": None,  # e.g. http://kafka-connect
    # `requests.auth.AuthBase` instance or tuple of (username, password) for Basic Auth
    "CONNECT_AUTH": None,
    # kwargs for `urllib3.util.retry.Retry` initialization
    "CONNECT_RETRY": dict(
        connect=5,
        read=5,
        status=5,
        backoff_factor=0.5,
        status_forcelist=[502, 503, 504],
    ),
    # `django_kafka.connect.client.KafkaConnectSession` would pass this value to every request method call
    "CONNECT_REQUESTS_TIMEOUT": 30,
    "CONNECTOR_NAME_PREFIX": "",
}


class Settings:
    @property
    def _settings(self):
        return getattr(django_settings, SETTINGS_KEY, {})

    def __getattr__(self, attr):
        if attr in self._settings:
            return self._settings[attr]

        if attr in DEFAULTS:
            return DEFAULTS[attr]

        raise AttributeError(f"Invalid setting: '{attr}'")


settings = Settings()



from unittest import mock

from django.test import TestCase

from django_kafka.retry.settings import RetrySettings
from django_kafka.topic import Topic


class RetrySettingTestCase(TestCase):
    def test_should_retry__include(self):
        settings = RetrySettings(max_retries=5, delay=60, include=[ValueError])

        self.assertEqual(settings.should_retry(ValueError()), True)
        self.assertEqual(settings.should_retry(IndexError()), False)

    def test_should_retry__exclude(self):
        settings = RetrySettings(max_retries=5, delay=60, exclude=[ValueError])

        self.assertEqual(settings.should_retry(ValueError()), False)
        self.assertEqual(settings.should_retry(IndexError()), True)

    def test_should_retry__no_include_exclude(self):
        settings = RetrySettings(max_retries=5, delay=60)

        self.assertEqual(settings.should_retry(ValueError()), True)

    def test_attempts_exceeded(self):
        settings = RetrySettings(max_retries=5, delay=60)

        self.assertEqual(settings.attempts_exceeded(6), True)
        self.assertEqual(settings.attempts_exceeded(5), False)
        self.assertEqual(settings.attempts_exceeded(4), False)

    @mock.patch("django.utils.timezone.now")
    def test_get_retry_time__fixed_delay(self, mock_now):
        settings = RetrySettings(max_retries=5, delay=60, backoff=False)

        timestamp = 1000
        mock_now.return_value.timestamp.return_value = timestamp
        self.assertEqual(settings.get_retry_timestamp(1), str(timestamp + 60))
        self.assertEqual(settings.get_retry_timestamp(2), str(timestamp + 60))
        self.assertEqual(settings.get_retry_timestamp(3), str(timestamp + 60))
        self.assertEqual(settings.get_retry_timestamp(4), str(timestamp + 60))

    @mock.patch("django.utils.timezone.now")
    def test_get_retry_time__backoff_delay(self, mock_now):
        settings = RetrySettings(max_retries=5, delay=60, backoff=True)

        timestamp = 1000
        mock_now.return_value.timestamp.return_value = timestamp
        self.assertEqual(settings.get_retry_timestamp(1), str(timestamp + 60))
        self.assertEqual(settings.get_retry_timestamp(2), str(timestamp + 120))
        self.assertEqual(settings.get_retry_timestamp(3), str(timestamp + 240))
        self.assertEqual(settings.get_retry_timestamp(4), str(timestamp + 480))


class RetryDecoratorTestCase(TestCase):
    def test_retry_decorator(self):
        class TopicA(Topic):
            pass

        retry = RetrySettings(max_retries=5, delay=60)
        result = retry(TopicA)

        self.assertIs(result, TopicA)
        self.assertIsNotNone(TopicA.retry_settings)
        self.assertEqual(
            [TopicA.retry_settings.max_retries, TopicA.retry_settings.delay],
            [5, 60],
        )



from confluent_kafka.serialization import Serializer


class NoOpSerializer(Serializer):
    """performs no messages serialization, returning the message without modification"""

    def __call__(self, obj, ctx=None):
        return obj



from django.test import TestCase
from django.utils import timezone

from django_kafka.retry.header import RetryHeader


class RetryHeaderTestCase(TestCase):
    def test_get_header(self):
        headers = [(RetryHeader.TIMESTAMP, "abc")]

        self.assertEqual(RetryHeader.get_header(headers, RetryHeader.TIMESTAMP), "abc")
        self.assertEqual(RetryHeader.get_header([], RetryHeader.TIMESTAMP), None)

    def test_get_retry_time(self):
        now = timezone.now()
        headers = [(RetryHeader.TIMESTAMP, str(now.timestamp()))]

        self.assertEqual(RetryHeader.get_retry_time(headers), now)

    def test_get_retry_time__none_when_unavailable(self):
        headers1 = [(RetryHeader.TIMESTAMP, "abc")]
        headers2 = []
        headers3 = None

        self.assertEqual(RetryHeader.get_retry_time(headers1), None)
        self.assertEqual(RetryHeader.get_retry_time(headers2), None)
        self.assertEqual(RetryHeader.get_retry_time(headers3), None)



from typing import Type
from unittest import mock
from unittest.mock import patch, Mock

from django.test import SimpleTestCase

from django_kafka.connect.connector import Connector
from django_kafka.consumer import Consumer, Topics
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.registry import ConsumersRegistry, Registry, ConnectorsRegistry


class RegistryTestCase(SimpleTestCase):
    def _gen_cls(self, name):
        return type(name, (object, ), {})

    def test_registering_same_key_not_allowed(self):
        registry = Registry()

        cls = self._gen_cls("SomeClass")
        key = registry.get_key(cls)
        error_msg = f"`{key}` is already registered."

        registry()(cls)

        with self.assertRaisesMessage(DjangoKafkaError, error_msg):
            registry()(cls)

    def test_get_key(self):
        cls_a = self._gen_cls("ClassA")
        cls_b = self._gen_cls("ClassB")
        registry = Registry()

        key_a = registry.get_key(cls_a)
        self.assertEqual(
            key_a,
            f"{cls_a.__module__}.{cls_a.__name__}",
        )

        key_b = registry.get_key(cls_b)
        self.assertEqual(
            key_b,
            f"{cls_b.__module__}.{cls_b.__name__}",
        )

    def test_decorator_adds_to_registry(self):
        cls_a = self._gen_cls("ClassA")
        cls_b = self._gen_cls("classB")

        registry = Registry()

        self.assertIs(registry()(cls_a), cls_a)
        self.assertIs(registry()(cls_b), cls_b)

        key_a = registry.get_key(cls_a)
        self.assertIs(registry[key_a], cls_a)

        key_b = registry.get_key(cls_b)
        self.assertIs(registry[key_b], cls_b)

    def test_iter_returns_expected_keys(self):
        cls_a = self._gen_cls("ClassA")
        cls_b = self._gen_cls("ClassB")
        registry = Registry()

        registry()(cls_a)
        registry()(cls_b)

        key_a = registry.get_key(cls_a)
        key_b = registry.get_key(cls_b)

        self.assertCountEqual(list(registry), [key_a, key_b])


class ConsumersRegistryTestCase(SimpleTestCase):
    def _get_consumer_cls(self, name, group_id) -> Type[Consumer]:
        return type[Consumer](
            name,
            (Consumer,),
            {"config": {"group.id": group_id}, "topics": Topics()},
        )

    @mock.patch("django_kafka.retry.consumer.RetryConsumer.build")
    def test_retry_consumer_registered(self, mock_retry_consumer_build):
        consumer_cls_a = self._get_consumer_cls("ConsumerA", "group_a")
        consumer_cls_b = self._get_consumer_cls("ConsumerB", "group_b")
        registry = ConsumersRegistry()
        retry_consumer_mock = mock.Mock()
        mock_retry_consumer_build.side_effect = [retry_consumer_mock, None]

        registry()(consumer_cls_a)
        registry()(consumer_cls_b)

        key_a = registry.get_key(consumer_cls_a)
        key_b = registry.get_key(consumer_cls_b)
        retry_key_a = f"{key_a}.retry"

        self.assertCountEqual(list(registry), [key_a, retry_key_a, key_b])
        self.assertIs(registry[retry_key_a], retry_consumer_mock)


@patch("django_kafka.connect.client.KafkaConnectSession", new=Mock())
class ConnectorsRegistryTestCase(SimpleTestCase):
    def _gen_cls(self, name) -> Type[Connector]:
        return type[Connector](
            name,
            (Connector,),
            {"config": {}},
        )

    @patch.multiple("django_kafka.conf.settings", CONNECT_HOST="http://kafka-connect")
    def test_get_key(self):
        connector_cls = self._gen_cls("ConnectorCls")
        registry = ConnectorsRegistry()

        key = registry.get_key(connector_cls)
        self.assertEqual(key, connector_cls().name)

    def test_registering_same_key_not_allowed(self):
        registry = ConnectorsRegistry()
        connector_a = self._gen_cls("ConnectorA")
        # give the same class name to the connector to generate same key
        connector_b = self._gen_cls("ConnectorA")

        key = registry.get_key(connector_a)
        error_msg = f"`{key}` is already registered."

        registry()(connector_a)

        with self.assertRaisesMessage(DjangoKafkaError, error_msg):
            registry()(connector_b)



import logging

from django.core.management import CommandError
from django.core.management.base import BaseCommand
from requests import Response
from requests.exceptions import RetryError

from django_kafka import kafka
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.connect.connector import Connector, ConnectorStatus
from django_kafka.utils import retry

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Publish connectors."

    def add_arguments(self, parser):
        parser.add_argument(
            "connector",
            type=str,
            default=None,
            nargs="?",
            help="Python path to the connector class(es). Processes all if not provided.",
        )
        parser.add_argument(
            "--list",
            action="store_true",
            default=False,
            help="List all connectors.",
        )
        parser.add_argument(
            "--validate",
            action="store_true",
            default=False,
            help="Validate connectors.",
        )
        parser.add_argument(
            "--publish",
            action="store_true",
            default=False,
            help="Create/Update/Delete connectors.",
        )
        parser.add_argument(
            "--check-status",
            action="store_true",
            default=False,
            help="Check status of connectors. Currently RUNNING status is considered as success.",
        )
        parser.add_argument(
            "--ignore-failures",
            action="store_true",
            default=False,
            help="The command wont fail if failures were detected. "
                 "By default if any failures were detected the command exist with error status.",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.connectors: list[str] = []
        self.has_failures = False

    def handle(self, connector, **options):
        if options["list"]:
            self.list_connectors()
            return

        if not any((connector, options["validate"], options["publish"], options["check_status"])):
            self.print_help("manage.py", "kafka_connect")
            return

        if connector:
            self.connectors = [connector]
        else:
            self.connectors = kafka.connectors

        if options["validate"]:
            self.handle_validate()

        if options["publish"]:
            self.handle_publish()

        if options["check_status"]:
            self.handle_status()

        if self.has_failures and not options["ignore_failures"]:
            raise CommandError("Command does not succeed.")

    def list_connectors(self):
        self.stdout.write(self.style.SUCCESS("Available connectors:"))
        for connector_path in kafka.connectors:
            if kafka.connectors[connector_path].mark_for_removal:
                self.stdout.write(f"- {connector_path} (marked for removal)")
            else:
                self.stdout.write(f"- {connector_path}")

    def handle_validate(self):
        self.stdout.write(self.style.SUCCESS("Validating connectors..."))
        
        for connector_path in self.connectors:
            self.stdout.write(f"{connector_path}: ", ending="")

            connector = kafka.connectors[connector_path]()

            if connector.mark_for_removal:
                self.stdout.write(self.style.WARNING("skip (REASON: marked for removal)"))
                continue

            try:
                connector.is_valid(raise_exception=True)
            except (DjangoKafkaError, RetryError) as error:
                self.has_failures = True
                self.stdout.write(self.style.ERROR("invalid"))
                self.stdout.write(self.style.ERROR(str(error)))
            else:
                self.stdout.write(self.style.SUCCESS("valid"))

    def handle_publish(self):
        self.stdout.write(self.style.SUCCESS("Publishing connectors..."))

        for connector_path in self.connectors:
            self.stdout.write(f"{connector_path}: ", ending="")

            connector = kafka.connectors[connector_path]()

            if connector.mark_for_removal:
                self.handle_delete(connector)
            else:
                self.handle_submit(connector)

    def handle_status(self):
        self.stdout.write(self.style.SUCCESS("Checking status..."))

        for connector_path in self.connectors:
            self.stdout.write(f"{connector_path}: ", ending="")

            connector = kafka.connectors[connector_path]()
            try:
                self._connector_is_running(connector)
            except DjangoKafkaError as error:
                self.has_failures = True
                self.stdout.write(self.style.ERROR(str(error)))

    @retry((DjangoKafkaError,), tries=3, delay=1, backoff=2)
    def _connector_is_running(self, connector: Connector):
        if connector.mark_for_removal:
            self.stdout.write(self.style.WARNING("skip (REASON: marked for removal)"))
            return
        try:
            status = connector.status()
        except DjangoKafkaError as error:
            if isinstance(error.context, Response) and error.context.status_code == 404:
                # retry: on 404 as some delays are expected
                raise

            self.has_failures = True
            self.stdout.write(self.style.ERROR("failed to retrieve"))
            self.stdout.write(self.style.ERROR(str(error)))
        except RetryError as error:
            self.has_failures = True
            self.stdout.write(self.style.ERROR("failed to retrieve"))
            self.stdout.write(self.style.ERROR(str(error)))
        else:
            if status == ConnectorStatus.RUNNING:
                self.stdout.write(self.style.SUCCESS(status))
            elif status == ConnectorStatus.UNASSIGNED:
                # retry: on UNASSIGNED as some delays are expected
                raise DjangoKafkaError(f"Connector status is {status}.")
            else:
                self.has_failures = True
                self.stdout.write(self.style.ERROR(status))

    def handle_delete(self, connector: Connector):
        try:
            deleted = connector.delete()
        except (DjangoKafkaError, RetryError) as error:
            self.has_failures = True
            self.stdout.write(self.style.ERROR("failed"))
            self.stdout.write(self.style.ERROR(str(error)))
        else:
            if deleted:
                self.stdout.write(self.style.SUCCESS("deleted"))
            else:
                self.stdout.write(self.style.WARNING("does not exist (already deleted)"))

    def handle_submit(self, connector: Connector):
        try:
            connector.submit()
        except (DjangoKafkaError, RetryError) as error:
            self.has_failures = True
            self.stdout.write(self.style.ERROR("failed"))
            self.stdout.write(self.style.ERROR(str(error)))
        else:
            self.stdout.write(self.style.SUCCESS("submitted"))



import logging
import traceback
from pydoc import locate
from typing import TYPE_CHECKING, Optional

from confluent_kafka import Consumer as ConfluentConsumer
from confluent_kafka import cimpl

from django_kafka.conf import settings
from django_kafka.exceptions import DjangoKafkaError

if TYPE_CHECKING:
    from django_kafka.topic import TopicConsumer

logger = logging.getLogger(__name__)


class Topics:
    _topic_consumers: tuple["TopicConsumer", ...]
    _match: dict[str, "TopicConsumer"]

    def __init__(self, *topic_consumers: "TopicConsumer"):
        self._topic_consumers = topic_consumers
        self._match: dict[str, "TopicConsumer"] = {}

    def get(self, topic_name: str) -> "TopicConsumer":
        if topic_name not in self._match:
            topic_consumer = next((t for t in self if t.matches(topic_name)), None)
            if not topic_consumer:
                raise DjangoKafkaError(f"No topic registered for `{topic_name}`")
            self._match[topic_name] = topic_consumer

        return self._match[topic_name]

    @property
    def names(self) -> list[str]:
        return [topic.name for topic in self]

    def __iter__(self):
        yield from self._topic_consumers


class Consumer:
    """
    Available settings of the producers (P) and consumers (C)
        https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
    Consumer configs
        https://kafka.apache.org/documentation/#consumerconfigs
    Kafka Client Configuration
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafka-client-configuration
    confluent_kafka.Consumer API
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-consumer
    """

    topics: Topics
    config: dict

    polling_freq = settings.POLLING_FREQUENCY
    default_logger = logger
    default_error_handler = settings.ERROR_HANDLER

    def __init__(self):
        self.config = self.build_config()
        self._consumer = ConfluentConsumer(self.config)

    def __getattr__(self, name):
        """proxy consumer methods."""
        return getattr(self._consumer, name)

    @classmethod
    def build_config(cls):
        return {
            "client.id": settings.CLIENT_ID,
            **settings.GLOBAL_CONFIG,
            **settings.CONSUMER_CONFIG,
            "logger": cls.default_logger,
            "error_cb": locate(cls.default_error_handler)(),
            **getattr(cls, "config", {}),
        }

    @property
    def group_id(self) -> str:
        return self.config["group.id"]

    def commit_offset(self, msg: cimpl.Message):
        if not self.config.get("enable.auto.offset.store"):
            # Store the offset associated with msg to a local cache.
            # Stored offsets are committed to Kafka by a background
            #  thread every 'auto.commit.interval.ms'.
            # Explicitly storing offsets after processing gives at-least once semantics.
            self.store_offsets(msg)

    def retry_msg(self, msg: cimpl.Message, exc: Exception) -> bool:
        from django_kafka.retry.topic import RetryTopicProducer

        topic_consumer = self.get_topic_consumer(msg)
        if not topic_consumer.retry_settings:
            return False

        return RetryTopicProducer(
            group_id=self.group_id,
            retry_settings=topic_consumer.retry_settings,
            msg=msg,
        ).retry(exc=exc)

    def dead_letter_msg(self, msg: cimpl.Message, exc: Exception):
        from django_kafka.dead_letter.topic import DeadLetterTopicProducer

        DeadLetterTopicProducer(group_id=self.group_id, msg=msg).produce_for(
            header_message=str(exc),
            header_detail=traceback.format_exc(),
        )

    def handle_exception(self, msg: cimpl.Message, exc: Exception):
        retried = self.retry_msg(msg, exc)
        if not retried:
            self.dead_letter_msg(msg, exc)
            self.log_error(exc)

    def get_topic_consumer(self, msg: cimpl.Message) -> "TopicConsumer":
        return self.topics.get(topic_name=msg.topic())

    def log_error(self, error):
        logger.error(error, exc_info=True)

    def process_message(self, msg: cimpl.Message):
        if msg_error := msg.error():
            self.log_error(msg_error)
            return

        try:
            self.get_topic_consumer(msg).consume(msg)
        # ruff: noqa: BLE001 (we do not want consumer to stop if message consumption fails in any circumstances)
        except Exception as exc:
            self.handle_exception(msg, exc)

        self.commit_offset(msg)

    def poll(self) -> Optional[cimpl.Message]:
        # poll for self.polling_freq seconds
        return self._consumer.poll(timeout=self.polling_freq)

    def start(self):
        self.subscribe(topics=self.topics.names)

    def run(self):
        try:
            self.start()
            while True:
                if (msg := self.poll()) is not None:
                    self.process_message(msg)
        except Exception as exc:
            self.log_error(exc)
            raise
        finally:
            self.stop()

    def stop(self):
        self.close()



from typing import Generic, TYPE_CHECKING, Type, TypeVar

from django_kafka.exceptions import DjangoKafkaError

if TYPE_CHECKING:
    from django_kafka.consumer import Consumer
    from django_kafka.connect.connector import Connector


T = TypeVar('T')


class Registry(Generic[T]):
    def __init__(self):
        self._classes: dict[str, Type[T]] = {}

    def __call__(self):
        def add_to_registry(cls: Type[T]) -> Type[T]:
            self.register(cls)
            return cls

        return add_to_registry

    def __getitem__(self, key: str):
        try:
            return self._classes[key]
        except KeyError as error:
            raise DjangoKafkaError(f"`{key}` is not registered.") from error

    def __iter__(self):
        yield from self._classes.keys()

    def get_key(self, cls: Type[T]) -> str:
        return f"{cls.__module__}.{cls.__name__}"

    def register(self, cls: Type[T]):
        key = self.get_key(cls)
        if key in self._classes:
            raise DjangoKafkaError(f"`{key}` is already registered.")
        self._classes[key] = cls


class ConnectorsRegistry(Registry["Connector"]):
    def get_key(self, cls) -> str:
        return cls.name


class ConsumersRegistry(Registry["Consumer"]):

    def register(self, cls):
        from django_kafka.retry.consumer import RetryConsumer
        
        super().register(cls)
        
        if retry_consumer_cls := RetryConsumer.build(cls):
            self._classes[f"{self.get_key(cls)}.retry"] = retry_consumer_cls



from unittest import mock

from django.test import TestCase, override_settings

from django_kafka.conf import SETTINGS_KEY
from django_kafka.dead_letter.header import DeadLetterHeader
from django_kafka.dead_letter.topic import DeadLetterTopicProducer


@override_settings(
    **{
        SETTINGS_KEY: {
            "RETRY_TOPIC_SUFFIX": "test-retry",
            "DEAD_LETTER_TOPIC_SUFFIX": "test-dlt",
        },
    },
)
class DeadLetterTopicProducerTestCase(TestCase):
    def test_name(self):
        mock_msg_topic_consumer = mock.Mock(**{"topic.return_value": "topic.name"})
        mock_msg_retry_topic = mock.Mock(
            **{"topic.return_value": "group.id.topic.name.test-retry.10"},
        )

        dlt_producer_1 = DeadLetterTopicProducer(
            group_id="group.id",
            msg=mock_msg_topic_consumer,
        )

        dlt_producer_2 = DeadLetterTopicProducer(
            group_id="group.id",
            msg=mock_msg_retry_topic,
        )

        self.assertEqual(dlt_producer_1.name, "group.id.topic.name.test-dlt")
        self.assertEqual(dlt_producer_2.name, "group.id.topic.name.test-dlt")

    def test_produce_for(self):
        msg_mock = mock.Mock(**{"topic.return_value": "msg_topic"})
        dlt_producer = DeadLetterTopicProducer(group_id="group.id", msg=msg_mock)
        dlt_producer.produce = mock.Mock()
        header_message = "header message"
        header_detail = "header detail"

        dlt_producer.produce_for(
            header_message=header_message,
            header_detail=header_detail,
        )

        dlt_producer.produce.assert_called_once_with(
            key=msg_mock.key(),
            value=msg_mock.value(),
            headers=[
                (DeadLetterHeader.MESSAGE, header_message),
                (DeadLetterHeader.DETAIL, header_detail),
            ],
        )



import contextlib
from abc import ABC, abstractmethod
from typing import Optional, Type

from confluent_kafka.serialization import MessageField
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model

from django_kafka.exceptions import DjangoKafkaError
from django_kafka.models import KafkaSkipModel
from django_kafka.topic import TopicConsumer


class ModelTopicConsumer(TopicConsumer, ABC):
    """Syncs abstract kafka messages directly in to Django model instances"""

    model: Optional[Type[Model]] = None  # override get_model for dynamic model lookups

    def get_defaults(self, model, value) -> dict:
        """
        Returns instance update_or_create defaults from the message value.

        value fields can be transformed by defining a transform_{attr} method.
        """
        defaults = {}
        for attr, attr_value in value.items():
            if transform_method := getattr(self, "transform_" + attr, None):
                new_attr, new_value = transform_method(model, attr, attr_value)
                defaults[new_attr] = new_value
            else:
                defaults[attr] = attr_value

        if issubclass(model, KafkaSkipModel):
            defaults["kafka_skip"] = True

        return defaults

    @abstractmethod
    def is_deletion(self, model, key, value) -> bool:
        """returns if the message represents a deletion"""

    @abstractmethod
    def get_lookup_kwargs(self, model, key, value) -> dict:
        """returns the lookup kwargs used for filtering the model instance"""

    def sync(self, model, key, value) -> tuple[Model, bool] | None:
        lookup = self.get_lookup_kwargs(model, key, value)

        if self.is_deletion(model, key, value):
            with contextlib.suppress(ObjectDoesNotExist):
                model.objects.get(**lookup).delete()
            return None

        defaults = self.get_defaults(model, value)
        return model.objects.update_or_create(**lookup, defaults=defaults)

    def get_model(self, key, value) -> Type[Model]:
        if self.model:
            return self.model
        raise DjangoKafkaError(
            "Cannot obtain model: either define a default model or override get_model",
        )

    def consume(self, msg):
        key = self.deserialize(msg.key(), MessageField.KEY, msg.headers())
        value = self.deserialize(msg.value(), MessageField.VALUE, msg.headers())
        model = self.get_model(key, value)
        self.sync(model, key, value)



import datetime

from django.test import TestCase

from django_kafka.topic.transforms import days_from_epoch_to_date


class TransformsTestCase(TestCase):
    def test_days_from_epoch_to_date(self):
        # days are calculated from epoch
        result = days_from_epoch_to_date(10)
        self.assertEqual(result, datetime.date(1970, 1, 11))

        # None values are silently skipped
        empty_values = [None, ""]
        default_value = "some-default"
        for value in empty_values:
            result = days_from_epoch_to_date(value, default=default_value)
            self.assertEqual(result, default_value)



import time
from functools import wraps
from typing import Type


def retry(
    exceptions: tuple[Type[Exception]] = (Exception,),
    tries: int = -1,
    delay: int = 0,
    backoff: int = 1,
):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            _tries = tries
            _delay = delay

            while _tries:
                try:
                    return f(*args, **kwargs)
                except exceptions:
                    _tries -= 1
                    if not _tries:
                        raise
                    time.sleep(_delay)
                    _delay *= backoff
        return wrapper
    return decorator



import re
from typing import TYPE_CHECKING

from django_kafka.conf import settings
from django_kafka.dead_letter.header import DeadLetterHeader
from django_kafka.retry.topic import RetryTopicProducer
from django_kafka.serialization import NoOpSerializer
from django_kafka.topic import TopicProducer

if TYPE_CHECKING:
    from confluent_kafka import cimpl


class DeadLetterTopicProducer(TopicProducer):
    key_serializer = NoOpSerializer
    value_serializer = NoOpSerializer

    def __init__(self, group_id: str, msg: "cimpl.Message"):
        self.group_id = group_id
        self.msg = msg
        super().__init__()

    @classmethod
    def suffix(cls):
        return settings.DEAD_LETTER_TOPIC_SUFFIX

    @property
    def name(self) -> str:
        topic = self.msg.topic()

        if re.search(RetryTopicProducer.pattern(), topic):
            return re.sub(RetryTopicProducer.pattern(), self.suffix(), topic)
        return f"{self.group_id}.{topic}.{self.suffix()}"

    def produce_for(self, header_message, header_detail):
        headers = [
            (DeadLetterHeader.MESSAGE, header_message),
            (DeadLetterHeader.DETAIL, header_detail),
        ]
        self.produce(
            key=self.msg.key(),
            value=self.msg.value(),
            headers=headers,
        )



from django.db import connection
from django.db.models import Model
from django.test import TestCase


class AbstractModelTestCase(TestCase):
    abstract_model: type[Model]
    model: type[Model]

    @classmethod
    def setUpClass(cls):
        class TestModel(cls.abstract_model):
            class Meta:
                app_label = "django_kafka"

        cls.model = TestModel

        with connection.schema_editor() as editor:
            editor.create_model(cls.model)

        # super has to be at the end, otherwise model won't initialize
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        with connection.schema_editor() as editor:
            editor.delete_model(cls.model)

        connection.close()



from unittest import mock

from django.db.models import Model
from django.test import TestCase

from django_kafka.exceptions import DjangoKafkaError
from django_kafka.models import KafkaSkipModel
from django_kafka.topic.model import ModelTopicConsumer


class TestModelTopicConsumer(TestCase):
    def _get_model_topic_consumer(self):
        class SomeModelTopicConsumer(ModelTopicConsumer):
            name = "name"
            model = Model

            def get_lookup_kwargs(self, model, key, value) -> dict:
                return {}

            def is_deletion(self, *args, **kwargs):
                return False

        return SomeModelTopicConsumer()

    def test_get_defaults(self):
        topic_consumer = self._get_model_topic_consumer()

        defaults = topic_consumer.get_defaults(model=Model, value={"name": 1})

        self.assertEqual(defaults, {"name": 1})

    def test_get_defaults__adds_kafka_skip(self):
        topic_consumer = self._get_model_topic_consumer()

        class KafkaSkip(KafkaSkipModel):
            pass

        defaults = topic_consumer.get_defaults(model=KafkaSkip, value={"name": 1})

        self.assertEqual(defaults, {"name": 1, "kafka_skip": True})

    def test_get_defaults__calls_transform_attr(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.transform_name = mock.Mock(return_value=("name_new", 2))

        defaults = topic_consumer.get_defaults(model=Model, value={"name": 1})

        topic_consumer.transform_name.assert_called_once_with(
            topic_consumer.model,
            "name",
            1,
        )
        self.assertEqual(defaults, {"name_new": 2})

    def test_sync(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.get_lookup_kwargs = mock.Mock(return_value={"id": "id"})
        topic_consumer.get_defaults = mock.Mock(return_value={"name": "name"})
        model = mock.Mock()

        results = topic_consumer.sync(model, {"key": "key"}, {"value": "value"})

        topic_consumer.get_lookup_kwargs.assert_called_once_with(
            model,
            {"key": "key"},
            {"value": "value"},
        )
        topic_consumer.get_defaults.assert_called_once_with(
            model,
            {"value": "value"},
        )
        model.objects.update_or_create.assert_called_once_with(
            id="id",
            defaults={"name": "name"},
        )
        self.assertEqual(results, model.objects.update_or_create.return_value)

    def test_sync__deleted(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.get_lookup_kwargs = mock.Mock(return_value={"id": "id"})
        topic_consumer.get_defaults = mock.Mock()
        topic_consumer.is_deletion = mock.Mock(return_value=True)
        model = mock.Mock()

        results = topic_consumer.sync(model, {"key": "key"}, {"value": "value"})

        topic_consumer.get_lookup_kwargs.assert_called_once_with(
            model,
            {"key": "key"},
            {"value": "value"},
        )
        topic_consumer.get_defaults.assert_not_called()
        model.objects.get.assert_called_once_with(id="id")
        model.objects.get.return_value.delete.assert_called_once()
        model.objects.update_or_create.assert_not_called()
        self.assertIsNone(results)

    def test_get_model(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = mock.Mock()

        self.assertEqual(
            topic_consumer.get_model({}, {}),
            topic_consumer.model,
        )

    def test_get_model__raises_when_model_not_set(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = None

        with self.assertRaises(DjangoKafkaError):
            topic_consumer.get_model({}, {})

    def test_consume(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.get_model = mock.Mock()
        msg_key = {"key": "key"}
        msg_value = {"value": "value"}
        topic_consumer.deserialize = mock.Mock(side_effect=[msg_key, msg_value])
        topic_consumer.sync = mock.Mock()

        topic_consumer.consume(mock.Mock())

        topic_consumer.get_model.assert_called_once_with(msg_key, msg_value)
        topic_consumer.sync.assert_called_once_with(
            topic_consumer.get_model.return_value,
            msg_key,
            msg_value,
        )



import logging
from pydoc import locate
from typing import Optional

from confluent_kafka import Producer as ConfluentProducer

from django_kafka.conf import settings

logger = logging.getLogger(__name__)


class Producer:
    """
    Available settings of the producers (P) and consumers (C):
        https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
    Producer configs
        https://kafka.apache.org/documentation/#producerconfigs
    Kafka Client Configuration
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafka-client-configuration
    confluent_kafka.Producer API
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-producer
    """

    config: dict

    default_logger = logger
    default_error_handler = settings.ERROR_HANDLER

    def __init__(self, config: Optional[dict] = None, **kwargs):
        kwargs.setdefault("logger", self.default_logger)
        kwargs.setdefault("error_cb", locate(self.default_error_handler)())

        self._producer = ConfluentProducer(
            {
                "client.id": settings.CLIENT_ID,
                **settings.GLOBAL_CONFIG,
                **settings.PRODUCER_CONFIG,
                **getattr(self, "config", {}),
                **(config or {}),
            },
            **kwargs,
        )

    def __getattr__(self, name):
        """
        proxy producer methods.
        """
        if name not in {"config"}:
            # For cases when `Producer.config` is not set and
            #  `getattr(self, "config", {})` is called on `__init__`,
            #  the initialization will fail because `_consumer` is not yet set.
            return getattr(self._producer, name)
        raise AttributeError(f"'{self.__class__.__name__}' has no attribute 'name'")



from unittest.mock import call, patch

from confluent_kafka.serialization import MessageField
from django.test import TestCase

from django_kafka.exceptions import DjangoKafkaError
from django_kafka.topic import TopicConsumer, TopicProducer


class SomeTopicProducer(TopicProducer):
    name = "some-topic-producer"


class TopicProducerTestCase(TestCase):
    def setUp(self):
        self.topic_producer = SomeTopicProducer()

    def test_init__disallows_regex(self):
        class RegexTopicProducer(TopicProducer):
            name = "^topic"

        with self.assertRaises(DjangoKafkaError):
            RegexTopicProducer()

    @patch("django_kafka.topic.TopicProducer.serialize")
    @patch("django_kafka.kafka.producer")
    def test_produce_serializer_kwargs(self, mock_kafka_producer, mock_topic_serialize):
        key = "key"
        value = "message value"
        headers = None  # default is None when not provided
        key_serializer_kwargs = {"a": "b"}
        value_serializer_kwargs = {"c": "d"}

        self.topic_producer.produce(
            value,
            key=key,
            key_serializer_kwargs=key_serializer_kwargs,
            value_serializer_kwargs=value_serializer_kwargs,
        )

        self.assertEqual(
            mock_topic_serialize.call_args_list,
            [
                call(
                    key,
                    MessageField.KEY,
                    headers,
                    **key_serializer_kwargs,
                ),
                call(
                    value,
                    MessageField.VALUE,
                    headers,
                    **value_serializer_kwargs,
                ),
            ],
        )

        mock_kafka_producer.produce.assert_called_once_with(
            self.topic_producer.name,
            mock_topic_serialize.return_value,
            key=mock_topic_serialize.return_value,
        )

    @patch("django_kafka.topic.TopicProducer.serialize")
    @patch("django_kafka.kafka.producer")
    def test_produce_only_value(self, mock_kafka_producer, mock_topic_serialize):
        value = "message value"
        headers = None  # default is None when not provided

        self.topic_producer.produce(value)

        mock_topic_serialize.assert_called_once_with(
            value,
            MessageField.VALUE,
            headers,
        )
        mock_kafka_producer.produce.assert_called_once_with(
            self.topic_producer.name,
            mock_topic_serialize.return_value,
        )

    @patch("django_kafka.topic.TopicProducer.serialize")
    @patch("django_kafka.kafka.producer")
    def test_produce_key_is_serialized(self, mock_kafka_producer, mock_topic_serialize):
        value = "message value"
        key = "some key"
        headers = {"header-1": "header-1-value"}

        self.topic_producer.produce(value, key=key, headers=headers)

        self.assertEqual(
            mock_topic_serialize.call_args_list,
            [
                call(key, MessageField.KEY, headers),
                call(value, MessageField.VALUE, headers),
            ],
        )

        mock_kafka_producer.produce.assert_called_once_with(
            self.topic_producer.name,
            mock_topic_serialize.return_value,
            key=mock_topic_serialize.return_value,
            headers=headers,
        )

    @patch("django_kafka.topic.TopicProducer.key_serializer")
    @patch("django_kafka.topic.TopicProducer.context")
    def test_serialize_key(self, mock_topic_context, mock_key_serializer):
        value = "some key"
        field = MessageField.KEY
        kwargs = {"key": "value"}

        self.topic_producer.serialize(value, field, **kwargs)

        mock_topic_context.assert_called_once_with(field, None)
        mock_key_serializer.assert_called_once_with(**kwargs)
        mock_key_serializer.return_value.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.TopicProducer.value_serializer")
    @patch("django_kafka.topic.TopicProducer.context")
    def test_serialize_value(self, mock_topic_context, mock_value_serializer):
        value = "some value"
        field = MessageField.VALUE
        kwargs = {"key": "value"}

        self.topic_producer.serialize(value, field, **kwargs)

        mock_topic_context.assert_called_once_with(field, None)
        mock_value_serializer.assert_called_once_with(**kwargs)
        mock_value_serializer.return_value.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.TopicProducer.key_serializer")
    @patch("django_kafka.topic.TopicProducer.value_serializer")
    @patch("django_kafka.topic.TopicProducer.context")
    def test_serialize_unknown_field(
        self,
        mock_topic_context,
        mock_value_serializer,
        mock_key_serializer,
    ):
        field = "something_unknown"

        with self.assertRaisesMessage(
            DjangoKafkaError,
            f"Unsupported serialization field {field}.",
        ):
            self.topic_producer.serialize("some value", field)

        mock_topic_context.assert_not_called()
        mock_value_serializer.assert_not_called()
        mock_key_serializer.assert_not_called()

    def test_context(self):
        fields = [MessageField.VALUE, MessageField.KEY]
        headers = {"header-1": "header-1-value"}

        for field in fields:
            with patch(
                "django_kafka.topic.SerializationContext",
            ) as mock_serialization_context:
                self.topic_producer.context(field, headers)
                mock_serialization_context.assert_called_once_with(
                    self.topic_producer.name,
                    field,
                    headers=headers,
                )


class SomeTopicConsumer(TopicConsumer):
    name = "some-topic-consumer"

    def consume(self, msg):
        pass


class TopicConsumerTestCase(TestCase):
    def setUp(self):
        self.topic_consumer = SomeTopicConsumer()

    def test_is_regex(self):
        self.topic_consumer.name = "^some-topic"

        self.assertTrue(self.topic_consumer.is_regex())

    def test_matches(self):
        self.topic_consumer.name = "some-topic"

        self.assertTrue(self.topic_consumer.matches("some-topic"))
        self.assertFalse(self.topic_consumer.matches("some-topic.extra"))

    def test_matches__regex(self):
        self.topic_consumer.name = "^some-topic"

        self.assertTrue(self.topic_consumer.matches("some-topic"))
        self.assertTrue(self.topic_consumer.matches("some-topic.extra"))
        self.assertFalse(self.topic_consumer.matches("other-topic"))

    @patch("django_kafka.topic.TopicConsumer.key_deserializer")
    @patch("django_kafka.topic.TopicConsumer.context")
    def test_deserialize_key(self, mock_topic_context, mock_key_deserializer):
        value = b"some key"
        field = MessageField.KEY
        kwargs = {"key": "value"}

        self.topic_consumer.deserialize(value, field, **kwargs)

        mock_topic_context.assert_called_once_with(field, None)
        mock_key_deserializer.assert_called_once_with(**kwargs)
        mock_key_deserializer.return_value.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.TopicConsumer.value_deserializer")
    @patch("django_kafka.topic.TopicConsumer.context")
    def test_deserialize_value(self, mock_topic_context, mock_value_deserializer):
        value = b"some value"
        field = MessageField.VALUE
        kwargs = {"key": "value"}

        self.topic_consumer.deserialize(value, field, **kwargs)

        mock_topic_context.assert_called_once_with(field, None)
        mock_value_deserializer.assert_called_once_with(**kwargs)
        mock_value_deserializer.return_value.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.TopicConsumer.key_deserializer")
    @patch("django_kafka.topic.TopicConsumer.value_deserializer")
    @patch("django_kafka.topic.TopicConsumer.context")
    def test_deserialize_unknown_field(
        self,
        mock_topic_context,
        mock_value_deserializer,
        mock_key_deserializer,
    ):
        field = "something_unknown"

        with self.assertRaisesMessage(
            DjangoKafkaError,
            f"Unsupported deserialization field {field}.",
        ):
            self.topic_consumer.deserialize("some value", field)

        mock_topic_context.assert_not_called()
        mock_value_deserializer.assert_not_called()
        mock_key_deserializer.assert_not_called()

    def test_context(self):
        fields = [MessageField.VALUE, MessageField.KEY]
        headers = {"header-1": "header-1-value"}

        for field in fields:
            with patch(
                "django_kafka.topic.SerializationContext",
            ) as mock_serialization_context:
                self.topic_consumer.context(field, headers)
                mock_serialization_context.assert_called_once_with(
                    self.topic_consumer.name,
                    field,
                    headers=headers,
                )
